;;; This code allows to translate a Prodigy problem description into PDDL,
;;; solve it with FF, and then follow the behaviour of FF during Prodigy search

(in-package "USER")

;; Translates a problem from Prodigy into PDDL, and solves it with FF
;; Problem is a pathname, a string (denoting a pathname), or a list
;; Returns the solutions to the problem
;; Side-effect: sets the value of *ff-solution*
;; Example: (generate-ff-solution 'depots "~/prodigy/domains/depots/probs/p2.lisp")
(defun generate-ff-solution (domain problem)
  (declare (special *ff-solution*))
  (multiple-value-bind (output-directory output-file)
      (translate-problem problem :domain domain)
    (execute-ff output-directory :file-prefix output-file)
    (setf *ff-solution* (car (solutions-ff)))
    (setf *prodigy-ff-solution* nil)
    (setf *ops-in-ff-solution* nil)
    (dolist (op *ff-solution* *ops-in-ff-solution*)
      (pushnew (car op) *ops-in-ff-solution* :test #'eq))))

;; Allows to follow the solution produced by FF
(defun follow-ff (problem)
  (setf *prodigy-ff-solution* nil)
  (setf *ops-in-ff-solution* nil)
  (dolist (op *ff-solution*)
    (pushnew (car op) *ops-in-ff-solution* :test #'eq))
  (load-domain)
  (my-load-problem problem)
  (run :output-level 3 :depth-bound (+ 4 (* 4 (1+ (length *ff-solution*))))))

#|

Meta-predicates

|#

(defun first-of-solution (operator)
  (let ((next-op (caar (or *prodigy-ff-solution*
			   (convert-solution *ff-solution*)))))
    (if (p4::strong-is-var-p operator)
	(list (list (cons operator next-op)))
      (eq (caar (or *prodigy-ff-solution*
		    (convert-solution *ff-solution*)))
	  operator))))
	       
;; Requires to set *prodigy-ff-solution* to nil at each run, and *ff-solution*
;;   to hold the ff-solution
(defun member-of-solution (instantiated-op)
  (and (member instantiated-op *prodigy-ff-solution* :test #'equal) t))

;; Requires to set *prodigy-ff-solution* to nil at each run, and *ff-solution*
;; to hold the ff-solution
(defun op-member-of-solution (op)
  (if (p4::strong-is-var-p op)
      (mapcar #'(lambda (op-in-solution)
		  (list (cons op op-in-solution)))
	      *ops-in-ff-solution*)
    (and (member op *prodigy-ff-solution* :test #'eq :key #'car) t)))

;; Version that checks only for the next op in solution
;; ;; Requires to set *prodigy-ff-solution* to nil at each run, and *ff-solution* to hold the ff-solution
;; (defun member-of-solution (instantiated-op)
;;   (when (member instantiated-op *prodigy-ff-solution* :test #'equal)
;;     ;; this is emptying the caching made at candidate-operator meta-predicate
;;     (makunbound '*candidate-operators*)
;;     (format t "~%Prodigy-FF-solution: ~a" *prodigy-ff-solution*)
;;     (pop *prodigy-ff-solution*)
;;     (break)
;;     t))

(defun convert-solution (solution)
  (setf *prodigy-ff-solution*
    (nreverse (mapcar #'(lambda (op)
			  (cons (car op)
				(mapcar #'(lambda (cte)
					    (p4::object-name-to-object cte *current-problem-space*))
					(cdr op))))
		      solution))))

;; goal is a literal, op is an op-name
(defun goal-effect-of-op-p (goal op)
  (let* ((op-struct (p4::rule-name-to-rule op *current-problem-space*))
	 (effects-list (car (if (negatedp goal)
				(p4::rule-del-list op-struct)
			      (p4::rule-add-list op-struct))))
	 (goal-name (car (get-pos-assertion goal))))
    (and (member goal effects-list
		 :test #'(lambda (goal effect)
			   (declare (ignore goal))
			   (eq goal-name (car (p4::effect-cond-effect effect)))))
	 t)))

;; (defun member-of-solution (instantiated-op)
;;   (mapcar #'(lambda (op-in-solution)
;; 	      (mapcar #'(lambda (x y)
;; 			  (cons x (p4::object-name-to-object y *current-problem-space*)))
;; 		      (cdr instantiated-op)
;; 		      (cdr op-in-solution)))
;; 	  (remove-if-not #'(lambda (op-in-solution)
;; 			     (eq (car instantiated-op) (car op-in-solution)))
;; 			 *ff-solution*)))

(defun candidate-operator-follow-ff (op)
  "Tests whether <op> is a member of the relevant ops being considered
at the current node.  Can be used as a generator."
  (declare (special *current-node* *candidate-operators*
		    *current-problem-space*))
  (let ((candidate-operators (if (boundp '*candidate-operators*)
				 *candidate-operators*
			       (p4::goal-node-ops-left
				(give-me-goal-node *current-node*)))))
    (and (member (p4::rule-name-to-rule op *current-problem-space*)
		 (if (eq candidate-operators :not-computed)
		     (setf *candidate-operators*
		       (p4::get-all-ops (p4::goal-node-goal *current-node*)
					*current-node*))
		   candidate-operators))
	 t)))

;;; The difference wrt target-goal is that it can generate alternatives if
;;; goal is a variable (i.e. can be used as a generator for the whole goal)
(DEFUN a-target-goal (goal)
  (declare (special *current-node* *current-problem-space*))
  (if (p4::strong-is-var-p goal)
      (mapcar #'(lambda (pending-goal)
		  (list (cons goal pending-goal)))
	      (get-my-pending-goals *current-node* t))
    (let ((bindings-prior-goals
	   (my-match-candidate-goal goal
				    (get-my-pending-goals *current-node* nil))))
      (setf (getf (p4::nexus-plist *current-node*) 'prior-goal)
	(cdr bindings-prior-goals))
      (car bindings-prior-goals))))

#| 

Control rules

;; Domain independent
(control-rule follow-goal
	      (if (and (a-target-goal <goal>)
		       (first-of-solution <op>)
		       (goal-effect-of-op-p <goal> <op>)))
	      (then select goal <goal>))

(control-rule follow-operator-behavior
	      (if (and (op-member-of-solution <op>)))
	      (then select operators <op>))
;;  (candidate-operator <op>)

;; Domain dependent (can be automatically generated)
(control-rule follow-bindings-drive-behavior
	      (if (and (current-operator drive)
		       (member-of-solution (drive <h> <c> <t>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>))))

(control-rule follow-bindings-lift-behavior
	      (if (and (current-operator lift)
		       (member-of-solution (lift <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))

(control-rule follow-bindings-drop-behavior
	      (if (and (current-operator drop)
		       (member-of-solution (drop <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))

(control-rule follow-bindings-load-behavior
	      (if (and (current-operator load)
		       (member-of-solution (load <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))

(control-rule follow-bindings-unload-behavior
	      (if (and (current-operator unload)
		       (member-of-solution (unload <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))
|#

#|
;; Strong version does not work: it requires that operators have the same order
;; as in the solution, but it does not have to be like that

(control-rule select-right-goal
	      (if (and (a-target-goal <goal>)
		       (first-of-solution <op>)
		       (goal-effect-of-op-p <goal> <op>)))
	      (then select goal <goal>))

(control-rule follow-drive-behavior
	      (if (and (candidate-operator drive)
		       (first-of-solution drive)))
	      (then select operator drive))

(control-rule follow-bindings-drive-behavior
	      (if (and (current-operator drive)
		       (member-of-solution (drive <h> <c> <t>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>))))

(control-rule follow-lift-behavior
	      (if (and (candidate-operator lift)
		       (first-of-solution lift)))
	      (then select operator lift))

(control-rule follow-bindings-lift-behavior
	      (if (and (current-operator lift)
		       (member-of-solution (lift <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))

(control-rule follow-drop-behavior
	      (if (and (candidate-operator drop)
		       (first-of-solution drop)))
	      (then select operator drop))

(control-rule follow-bindings-drop-behavior
	      (if (and (current-operator drop)
		       (member-of-solution (drop <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))

(control-rule follow-load-behavior
	      (if (and (candidate-operator load)
		       (first-of-solution load)))
	      (then select operator load))

(control-rule follow-bindings-load-behavior
	      (if (and (current-operator load)
		       (member-of-solution (load <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))

(control-rule follow-unload-behavior
	      (if (and (candidate-operator unload)
		       (first-of-solution unload)))
	      (then select operator unload))

(control-rule follow-bindings-unload-behavior
	      (if (and (current-operator unload)
		       (member-of-solution (unload <h> <c> <t> <d>))))
	      (then select bindings
		    ((<x> . <h>) (<y> . <c>) (<z> . <t>) (<p> . <d>))))
|#