;; Copyright (c) 2004 James Bailey (dgym.REMOVE_THIS.bailey@gmail.com).
;;
;; Permission is hereby granted, free of charge, to any person obtaining a
;; copy of this software and associated documentation files (the "Software"), to
;; deal in the Software without restriction, including without limitation the
;; rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
;; sell copies of the Software, and to permit persons to whom the Software is
;; furnished to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice shall be included in all
;; copies or substantial portions of the Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
;; OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
;; SOFTWARE.

(defpackage :gtk
  (:use :common-lisp :ffi))

(in-package :gtk)

#+unix
(eval-when (:compile-toplevel eval)

  (defconstant *gtk-lib* "libgtk-x11-2.0.so.0")
  (defconstant *gdk-lib* *gtk-lib*)
  (defconstant *glib-lib* *gtk-lib*)
  (defconstant *gthread-lib* "libgthread-2.0.so.0")
  (defconstant *gobj-lib* *gtk-lib*))
#+win32
(eval-when (:compile-toplevel eval)
  (defconstant *gtk-lib* "libgtk-win32-2.0-0.dll")
  (defconstant *gdk-lib* "libgdk-win32-2.0-0.dll")
  (defconstant *glib-lib* "libglib-2.0-0.dll")
  (defconstant *gobj-lib* "libgobject-2.0-0.dll"))

(def-c-type widget c-pointer)

;;;
;; import macros
;;;

(defmacro trap-errors (err-val &rest forms)
  `(catch 'trap-errors
          (handler-bind ((error (lambda (arg) 
                                  (declare (ignore arg))
                                  (throw 'trap-errors ,err-val))))
            ,@forms)))

(defmacro make-imports ()
  (cons 'progn
        (loop for (name lib) in `((gtk-import ,*gtk-lib*)
                                  (gdk-import ,*gdk-lib*)
                                  (glib-import ,*glib-lib*)
                                  (gobj-import ,*gobj-lib*)
                                  (gthread-import ,*gthread-lib*))
              collect
              `(defmacro ,name (ret-type name &rest args)
                 (let ((sym (if (consp name)
                              (car name)
                              (read-from-string name))))

                   `(trap-errors
                      nil
                      (progn
                        (ffi:def-call-out ,sym
                                          (:name
                                            ,(if (consp name)
                                               (second name)
                                               name))
                                          (:library ,,lib)
                                          (:arguments ,@args)
                                          (:return-type ,ret-type)
                                          (:language :stdc)))
                      (export ',sym)))))))

(make-imports)

;;;
;; callback handling
;;;

;; this is complicated as each callback can have a different number of arguments
;; and it is the last that is most interesting (a gpointer to some data we manage)

;; in order to call lisp functions (and thus support closures) rather than being limited
;; to a primitive data type, the data parameter is used by this module, as an index into
;; a global vector of callback functions.
;; so, a call is made to gtk_connect, with an object, the event name and a callback function
;; the callback is put in the global vector, and its index is given as the data argument to
;; g_signal_connect_data. This module's callback function performs the lookup (it gets the
;; index as its last parameter), and calls it.

;; this could be handled directly by the ffi (passing in a lisp function as the callback would work)
;; but this causes memory leaks.

;;;
;; make the connect functions, taking from 0-4 arguments as well as the object
;;;

(defmacro make-connect-funcs (param-count)
  (let* ((c-name (intern (format nil "G_SIGNAL_CONNECT_DATA-~a" param-count)))
         (cb-name (intern (format nil "GTK-CALLBACK-~a" param-count)))
         (args
           (loop for i from 1 upto param-count
                 collect (intern (format nil "ARG-~a" i))))
         (arg-types
           (mapcar (lambda (x) `(,x c-pointer)) args)))
    `(progn
       (gobj-import int (,c-name "g_signal_connect_data")
                    (widget widget)
                    (name c-string)
                    ;; callback
                    (callback
                      (c-function (:arguments
                                    (object c-pointer)
                                    ,@arg-types
                                    (data int))
                                  (:return-type int)
                                  (:language :stdc)))
                    (data int)
                    ;; clean up
                    (clean-up
                      (c-function (:arguments
                                    (data int))
                                  (:return-type nil)
                                  (:language :stdc)))
                    (b int))

       (ffi:def-call-in ,(intern (format nil "GTK-CALLBACK-~a" param-count))
                        (:arguments
                          (object c-pointer)
                          ,@arg-types
                          (data ffi:c-string))
                        (:return-type ffi:int)
                        (:language :stdc))

       (defun ,(intern (format nil "GTK-CALLBACK-~a" param-count)) (object ,@args data)
         (funcall (aref *callback-funcs* data)
                  object
                  ,@args))

       (setf (aref *connect-funcs* ,param-count) (cons #',c-name #',cb-name)))))

(defvar *connect-funcs* (make-array 5))
(defvar *callback-funcs* (make-array 0 :adjustable t :fill-pointer 0 :element-type t))
(make-connect-funcs 0)
(make-connect-funcs 1)
(make-connect-funcs 2)
(make-connect-funcs 3)
(make-connect-funcs 4)

;;;
;; our cleanup function just has to discard the entry in *callback-funcs*
;;;

(ffi:def-call-in gtk-cleanup (:arguments (data int))
                 (:return-type nil)
                 (:language :stdc))

(defun gtk-cleanup (data)
  (setf (aref *callback-funcs* data) nil)
  nil)

;;;
;; These struct definitions allow us to get the itype from an object
;;;

(def-c-struct GTypeClass (g_type int))
(def-c-struct GTypeInstance (g_class (c-ptr GTypeClass)))
(def-c-struct GSignalQuery
              (signal_id int)
              (signal_name c-string)
              (itype int)
              (signal_flags int)
              (return_type int)
              (n_params int)
              (param_types c-pointer))


;;Estas estructuras son mias jd
;;typedef struct {
;;  guint32 pixel;
;;  guint16 red;
;;  guint16 green;
;;  guint16 blue;
;;} GdkColor;

(def-c-struct GdkColor
              (pixel long)
	     (red int)
	     (green int)
	     (blue int))

;;typedef struct {
;;  gpointer data;
;;  GList *next;
;;  GList *prev;
;;} GList;
(def-c-struct GList
              (data c-pointer)
	     (next (c-ptr GList))
	     (prev(c-ptr GList)))

;;struct GdkWindow
;;  {
;;    gpointer user_data;
;;  };
(def-c-struct GdkWindow
              (user_data c-pointer))

;;typedef struct {
;;  GdkAxisUse use;
;;  gdouble    min;
;;  gdouble    max;
;;} GdkDeviceAxis;
(def-c-struct GdkDeviceAxis
              (use int)
	     (min double-float)
	     (max double-float))

;;typedef struct {
;;  guint keyval;
;;  GdkModifierType modifiers;
;;} GdkDeviceKey;
(def-c-struct GdkDeviceKey
              (keyval int)
              (modifiers int))

;;typedef struct {
;;  GObject parent_instance;
	  
;;  gchar *name;
;;  GdkInputSource source;
;;  GdkInputMode mode;
;;  gboolean has_cursor; /* TRUE if the X pointer follows device motion */
	  
;;  gint num_axes;
;;  GdkDeviceAxis *axes;
	  
;;  gint num_keys;
;;  GdkDeviceKey *keys;
;;} GdkDevice;

(def-c-struct GdkDevice
              (parent_instance c-pointer)
	      (name c-string)
	      (source int)
	      (mode int)
	      (has_cursor int)
	      (num_axes int)
	      (axes (c-ptr GdkDeviceAxis))
	      (num_keys int)
	      (keys (c-ptr GdkDeviceKey)))


;;typedef struct {
;;  GdkEventType type;
;;  GdkWindow *window;
;;  gint8 send_event;
;;  guint32 time;
;;  gdouble x;
;;  gdouble y;
;;  gdouble *axes;
;;  guint state;
;;  gint16 is_hint;
;;  GdkDevice *device;
;;  gdouble x_root, y_root;
;;} GdkEventMotion;
(def-c-struct GdkEventMotion
              (type int)
	      (window c-pointer)
	      (send_event int)
	      (time long)
	      (x double-float)
	      (y double-float)
	      (axes (c-ptr double-float))
	      (state int)
	      (is_hint long)
	      (device (c-ptr GdkDevice))
	      (x_root double-float)
	      (y_root double-float))
;; fin jd



(gobj-import nil "g_signal_handler_disconnect" (obj widget) (id int))
(gobj-import int "g_signal_lookup" (name c-string) (itype int))
(gobj-import nil "g_signal_query" (id int) (query c-pointer))
(gobj-import int "g_type_from_name" (name c-string))


(defun get_type_from_instance (widget)
  ;; returns the type from an instance
  ;; instance->g_class->g_type
  (with-c-var (_widget 'c-pointer widget)
              (slot
                (deref
                  (slot (deref (cast _widget '(c-ptr GTypeInstance)))
                        'g_class))
                'g_type)))

;;;
;; The exported function, gtk_connect, taking a gobject, a
;; signal name (e.g "delete_event") and a callback function
;; The callback function will be passed the gobject, and any other
;; signal specific parameters, but not a data parameter.
;;;
(export 'gtk_connect)
(defun gtk_connect (widget signal func)
  (let* ((n_params
           (with-c-var (query 'GSignalQuery)
                       (g_signal_query (g_signal_lookup signal (get_type_from_instance widget))
                                       (c-var-address query))
                       (slot query 'n_params)))
         (funcs (aref *connect-funcs* n_params))
         
         (idx (position nil *callback-funcs*)))

    (when (null idx)
      (setq idx (vector-push-extend func *callback-funcs*)))

    (funcall (car funcs) 
             widget
             signal
             (cdr funcs)
             idx
             #'gtk-cleanup
             0)))


;;;
;; the actual imports
;;;

;; rather than coding these in by hand, they are read from gtk-server.cfg
;; this file is part of the (excellent) http://www.gtk-server.org project,
;; and defines lots of gtk functions in a simple-enough-to-parse form.
;; 
;; the scanning is done in a macro, so it is performed at compile time. there
;; is no need to ship gtk-server.cfg with you project.

(defmacro read-gtk-server-cfg ()
  (declare (special *gtk-server-file*))
  (labels ((convert-type (typ)
                         (let ((sym (read-from-string typ)))
                           (case sym
                             (none 'nil)
                             (long 'int)
                             (bool 'int)
                             (string 'c-string)
                             (float 'single-float)
                             (double 'double-float)
                             (null 'c-pointer)
                             (otherwise sym))))

           (proc-line (string)

                      (let ((parts (loop for i = 0 then (1+ j)
                                         as j = (position #\, string :start i)
                                         collect (string-trim " " (subseq string i j))
                                         while j)))

                        ;; parts are: API name, callback signal type, return value, amount of arguments, arg1, arg2...
                        `(gtk-import ,(convert-type (third parts))
                                     ,(first parts)
                                     ,@(loop for arg in (nthcdr 4 parts)
                                             collect `(arg ,(convert-type arg)))))))

    `(progn
       ,@(with-open-file (cfg user::*gtk-server-file*)
           (loop with forms = nil
                 finally (return forms)
                 for line = (read-line cfg nil)
                 while line
                 do
                 (setq line (string-trim (coerce '(#\space #\tab #\newline #\return) 'string) line))
                 (when (and (plusp (length line))
                            (not (char= (aref line 0) #\#)))

                   (with-input-from-string (*standard-input* line)
                     ;; check it starts with GTK_LIB_FUNCTION
                     (when (eq (read) 'gtk_lib_function)
                       ;; skip the =
                       (read)
                       (push (proc-line (read-line)) forms)))))))))
 
(read-gtk-server-cfg)

(def-c-struct GtkTreeIter 
              (stamp int)
              (user_data c-pointer)
              (user_data2 c-pointer)
              (user_data3 c-pointer))

(def-c-struct GValue
              (g_type int)
              (unknown1 double-float)
              (unknown2 double-float)
              (unknown3 double-float)
              (unknown4 double-float))

(gobj-import int "g_type_fundamental" (val int))
(gobj-import nil "g_value_init" (val c-pointer) (gtype int))
(gobj-import nil "g_value_set_string" (val c-pointer) (str c-string))
(gobj-import nil "g_object_set" (object c-pointer) (property c-string) (a int) (c c-pointer)) 
(gobj-import nil "g_object_set_data" (obj c-pointer) (key c-string) (data c-pointer))
(gobj-import c-pointer "g_object_get_data" (obj c-pointer) (key c-string))
(gobj-import c-pointer "g_object_unref" (obj c-pointer))
;;(gobj-import nil "g_locale_to_utf8" (val c-pointer) (str c-string))
;;(gobj-import c-pointer "gtk_adjustment_new" (a double-float) (b double-float) (c double-float) 
;;(d double-float)(e double-float) (f double-float))

(gtk-import c-pointer "gtk_list_store_newv" (count int) (types (c-array-ptr int)))
(gtk-import nil "gtk_list_store_clear" (store c-pointer))
(gtk-import nil "gtk_list_store_append" (store c-pointer) (iter (c-ptr GtkTreeIter) :out))
(gtk-import nil "gtk_list_store_set_value" (store c-pointer) (iter (c-ptr GtkTreeIter)) (column int) (val c-pointer))
(gtk-import nil "gtk_list_store_set" (store c-pointer) (iter (c-ptr GtkTreeIter)) (column int) (val c-string) (terminator int))


;;Esto esta anadido por jd
;;gboolean gtk_tree_store_remove (GtkTreeStore *tree_store,GtkTreeIter *iter);
(gtk-import int  "gtk_tree_store_remove" (model c-pointer) (iter (c-ptr GtkTreeIter) :in-out))

;;void gtk_list_unselect_all(GtkList *list);
(gtk-import nil  "gtk_list_unselect_all" (widget widget))

;;void gtk_tree_store_insert (GtkTreeStore *tree_store,GtkTreeIter *iter,GtkTreeIter *parent,gint position);
(gtk-import nil  "gtk_tree_store_insert" (model c-pointer) (toplevel (c-ptr GtkTreeIter) :out) (parent (c-ptr-null GtkTreeIter)) (position int)) 

;;gboolean gtk_tree_model_iter_next(GtkTreeModel *tree_model,GtkTreeIter *iter);
(gtk-import int  "gtk_tree_model_iter_next" (model c-pointer) (iter (c-ptr GtkTreeIter) :in-out))


;;gboolean gtk_tree_model_get_iter_first(GtkTreeModel *tree_model,GtkTreeIter *iter);
(gtk-import int  "gtk_tree_model_get_iter_first" (model c-pointer) (iter (c-ptr GtkTreeIter) :out))

;;gboolean gtk_tree_model_iter_children(GtkTreeModel *tree_model,GtkTreeIter *iter,GtkTreeIter *parent);
(gtk-import int  "gtk_tree_model_iter_children" (model c-pointer) (iter_child (c-ptr GtkTreeIter) :out) (iter_father (c-ptr GtkTreeIter):in))


;;GtkTreeSelection* gtk_tree_view_get_selection(GtkTreeView *tree_view)
;;(def-c-struct GtkTreeSelection)
(gtk-import c-pointer  "gtk_tree_view_get_selection" (widget widget))


;;gboolean gtk_tree_selection_get_selected (GtkTreeSelection *selection,GtkTreeModel **model,GtkTreeIter *iter);
(gtk-import int "gtk_tree_selection_get_selected" (widget widget) (model c-pointer) (iter (c-ptr GtkTreeIter) :out))

;;void gtk_tree_model_get(GtkTreeModel *tree_model,GtkTreeIter *iter,...);
(gtk-import nil "gtk_tree_model_get" (model c-pointer) (iter (c-ptr GtkTreeIter) :in-out) (column int) (name (c-ptr c-string) :out) (terminator int))

;;GdkWindow*  gdk_window_get_pointer(GdkWindow *window,gint *x,gint *y,GdkModifierType *mask);
;; (gdk-import c-pointer "gdk_window_get_pointer" (model (c-ptr GdkWindow)) (x (c-ptr int):out) (y (c-ptr int):out) (mask (c-ptr int):out))
;;GdkWindow*  gdk_window_get_pointer(GdkWindow *window,gint *x,gint *y,GdkModifierType *mask);
(gdk-import c-pointer "gdk_window_get_pointer" (model c-pointer) (x (c-ptr int):out) (y (c-ptr int):out) (mask (c-ptr int):out))

;;void gtk_tree_selection_set_mode(GtkTreeSelection *selection,GtkSelectionMode type);
(gtk-import nil "gtk_tree_selection_set_mode" (widget c-pointer) (modo int))

;;GtkTreePath* gtk_tree_model_get_path(GtkTreeModel *tree_model,GtkTreeIter *iter);
(gtk-import c-pointer "gtk_tree_model_get_path" (model c-pointer) (iter (c-ptr GtkTreeIter)))

;;void gtk_tree_model_row_changed(GtkTreeModel *tree_model,GtkTreePath *path,GtkTreeIter *iter);
(gtk-import nil "gtk_tree_model_row_changed" (model c-pointer) (path c-pointer) (iter (c-ptr GtkTreeIter) :out))

;;GtkWidget*  gtk_list_new (void);
(gtk-import widget "gtk_list_new")

;;void  gtk_list_set_selection_mode(GtkList *list,GtkSelectionMode mode);
(gtk-import nil "gtk_list_set_selection_mode" (list widget) (mode int))

;;El gtk_signal_connect esta metido?

;;GtkWidget*  gtk_list_item_new_with_label(const gchar *label);
(gtk-import widget "gtk_list_item_new_with_label" (label c-string))


;;gchar* gtk_tree_path_to_string(GtkTreePath *path);
(gtk-import c-string "gtk_tree_path_to_string" (path c-pointer))

;;void g_free (gpointer mem);
(gobj-import nil "g_free" (mem c-pointer))

(gtk-import c-string "strcpy" (cad1 c-pointer) (cad2 c-pointer))
;;gboolean gtk_tree_model_iter_parent(GtkTreeModel*tree_model,GtkTreeIter *iter,GtkTreeIter *child);
(gtk-import int  "gtk_tree_model_iter_parent" (model c-pointer) (parent (c-ptr GtkTreeIter) :out) (child (c-ptr GtkTreeIter)))


;; fin jd



(gtk-import nil "gtk_tree_view_set_model" (widget widget) (model c-pointer))

(gtk-import c-pointer "gtk_cell_renderer_text_new")

(gtk-import c-pointer "gtk_tree_view_column_new")
(gtk-import c-pointer "gtk_tree_view_column_new_with_attributes" (title c-string) (renderer c-pointer) (name c-string) (value int) (terminator c-pointer))

(gtk-import int "gtk_tree_view_append_column" (view widget) (column c-pointer))
(gtk-import nil "gtk_tree_view_column_set_title" (view widget) (title c-string))
(gtk-import nil "gtk_tree_view_column_set_visible" (view widget) (value int))
(gtk-import nil "gtk_tree_view_column_set_spacing" (view widget) (value int))
(gtk-import nil "gtk_tree_view_column_pack_start" (column c-pointer) (renderer c-pointer) (expand int))
(gtk-import nil "gtk_tree_view_column_set_attributes" (column c-pointer) (renderer c-pointer) (name c-string) (value int) (terminator nil))
(gtk-import nil "gtk_tree_view_column_add_attribute" (column c-pointer) (renderer c-pointer) (name c-string) (value int))

;; raquel
(gtk-import c-pointer "gtk_tree_store_newv" (count int) (types (c-array-ptr int)))
;;(gtk-import nil "gtk_tree_store_append" (store c-pointer) (iter (c-ptr GtkTreeIter)) (iter (c-ptr GtkTreeIter))) 
(gtk-import nil "gtk_tree_store_append" (store c-pointer) (toplevel (c-ptr GtkTreeIter) :out) (parent (c-ptr-null GtkTreeIter))) 
(gtk-import nil "gtk_tree_store_set" (store c-pointer) (iter (c-ptr GtkTreeIter)) (column int) 
                                     ;;(val (c-ptr (c-array int 100))) 
                                     (val c-string)
                                     (terminator int))

;;raquel
;;gthread imports
(gthread-import nil "g_thread_init" (table c-pointer))
(gthread-import c-pointer "g_thread_create" (func c-string) (data c-pointer) (joinable int) (err c-pointer))
;;gdk imports
(gdk-import nil "gdk_threads_init") 
(gdk-import nil "gdk_threads_enter") 
(gdk-import nil "gdk_threads_leave")
;; fin raquel
;;;
;; memory leak test
;;;

#+nil
(defun ml-test ()
  (gtk_init 0 0)
  (let ((w (gtk_window_new 0)))
    (gtk_widget_show_all w)
    (loop for id = (gtk_connect w "delete_event" (lambda (&rest args) (print "destroyed") (ext:quit)))
          do
          (g_signal_handler_disconnect w id)
          (ext:gc)
          (print (room))
          (sleep 0.1))
    (gtk_main)))



