;;; Connection between IPSS and a path-planner. Requirements: having the search code from Daniel.
;;; Use: add the call to (path-planner init-node end-node quality-measure)
;;; within the definition of a variable in the preconds of an IPSS
;;; operator. 

;; This function will return NIL if no path, or the cost of the
;; solution in terms of quality-measure. It will also bind the global
;; variables: *pp-utility*, *pp-cost*, *pp-time*, *pp-distance*, *pp-other-metric*
;; pp-graph is used in domains (such as driverlog) where there are more than
;; one pp-graph.
(defun path-planner (init-node end-node &optional (quality-measure 'time)
						  (solution-function #'do-nothing) (pp-graph 'move))
  (declare (special *pp-utility* *pp-cost* *pp-time* *pp-distance* *pp-other-metric*))
  (let* ((init-node-name (convert-object-in-name init-node))
	 (end-node-name (convert-object-in-name end-node))
	 (cached-solution (search-cache init-node-name end-node-name pp-graph))
	 (solution (or cached-solution
		       (provide-solution init-node-name end-node-name quality-measure))))
    (if (not cached-solution)
	(cache-solution init-node-name end-node-name solution))
    (cond ((eq solution 'no-solution)
	   (setf *pp-utility* 0)
	   (setf *pp-cost* 0)
	   (setf *pp-time* 0)
	   (setf *pp-distance* 0)
	   (setf *pp-other-metric* 0)
	   nil)
	  (t (setf *pp-utility* (compute-pp-solution-quality solution 'utility))
	     (setf *pp-cost* (compute-pp-solution-quality solution 'cost))
	     (setf *pp-time* (compute-pp-solution-quality solution 'time))
	     (setf *pp-distance* (compute-pp-solution-quality solution 'distance))
	     (setf *pp-other-metric* (compute-pp-solution-quality solution 'other-metric))
	     (list (funcall solution-function (case quality-measure
						(utility *pp-utility*)
						(cost *pp-cost*)
						(time *pp-time*)
						(distance *pp-distance*)
						(t *pp-other-metric*))))))))

(defun compute-pp-solution-quality (solution quality-measure)
  (apply #'+ (mapcar #'(lambda (op)
			 (nth (case quality-measure
				(utility 4)
				(distance 5)
				(time 6)
				(price 7)
				(t 8))
			      op))
		     solution)))

(defun search-cache (pinit pf op-name)
  (declare (ignore op-name)
	   (special *solutions-hash*))
  (let ((pinit-sols (gethash pinit *solutions-hash*)))
    (if pinit-sols
	(cdr (assoc pf pinit-sols)))))

(defun cache-solution (pinit pf solution)
  (declare (special *solutions-hash*))
  (push (cons pf solution)
	(gethash pinit *solutions-hash*)))

(defun do-nothing (value)
  value)