;;;; Code to translate IPSS problem descriptions into PDDL

(in-package "USER")

;;; Translates problem descriptions from a probset file in IPSS into
;;; PDDL format
(defun translate-problems-to-pddl (file &key (domain 'logistics)
					(output-directory *ff-output-directory*)
					(metric-info nil))
  (declare (special *all-problems*))
  (load file)
  (dolist (problem *all-problems*)
      (translate-one-problem problem nil domain output-directory metric-info)))

;; It translates an individual problem from a single problem file
(defun translate-problem (problem &key (domain 'logistics)
				       (output-directory *ff-output-directory*)
				       (metric-info nil))
  (declare (special *ff-output-directory*))
  (cond ((or (stringp problem) (pathnamep problem))
	 (if (probe-file problem)
	     (with-open-file (istream problem :direction :input)
	       (translate-one-problem (read istream) (pathname-name problem)
				      domain output-directory metric-info))
	   (format t "~% File not found: ~a" problem)))
	((listp problem)
	 (translate-one-problem problem (cadadr problem)
				domain output-directory metric-info))
	(t (format t "~%Nothing to translate ~a" problem)
	   nil)))

(defun translate-one-problem (problem name domain output-directory metric-info)
  (let* ((problem-description (caddr problem))
	 (problem-name (cadadr problem-description))
	 (objects-defs (cdaddr problem-description))
	 (state (nth 3 problem-description))
	 (goals (nth 4 problem-description))
	 (output-file (format nil "~(~a-~a~).pddl" (or name problem-name) domain)))
    (with-open-file (ostream (format nil "~(~a~a~)" output-directory output-file)
		     :direction :output :if-exists :supersede
		     :if-does-not-exist :create)
      (format ostream "(define (problem ~(~a~)) (:domain ~(~a~))~%" problem-name domain)
      (write-pddl-objects objects-defs domain ostream)
      (write-pddl-state state domain ostream)
      (write-pddl-goals goals domain ostream)
      (if metric-info
	  (format ostream "  ~(~a~)" metric-info))
      (format ostream ")" problem-name domain))
    (values output-directory output-file)))

(defun write-pddl-objects (objects-defs domain ostream)
  (format ostream "  (:objects")
  (dolist (types objects-defs)
    (format ostream "~%    ")
    (let ((type (car (last types)))
	  (objects (butlast (if (eq (car types) 'objects-are)
				(cdr types)
			      types))))
      (dolist (object objects)
	(format ostream " ~(~a~)" object))
      (format ostream " - ~a" (change-names-to-pddl type domain))))
  (format ostream "  )~%"))

(defun write-pddl-state (state domain ostream)
  (format ostream "  (:init")
  (let ((the-number nil))
    (dolist (literal (change-names-to-pddl (cdadr state) domain))
      (cond ((setq the-number (find-if #'numberp literal))
	     (format ostream "~%    (= (")
	     (dolist (element literal)
	       (if (not (numberp element))
		   (format ostream "~(~a~) " element)))
	     (format ostream ") ~a)" the-number))
	    (t (format ostream "~%    ~(~a~)" literal))))
    (format ostream ")~%")))

(defun write-pddl-goals (goals domain ostream)
  (let* ((the-goals (cadr goals))
	 (more-than-one-goal-p  (eq (car the-goals) 'and)))
    (format ostream "  (:goal")
    (if more-than-one-goal-p
	(format ostream "    (and"))
    (dolist (literal (change-names-to-pddl (if more-than-one-goal-p
					       (cdr the-goals)
					       (cdr goals))
					   domain))
	(format ostream "~%      ~(~a~)" literal))
    (if more-than-one-goal-p
	(format ostream ")"))
    (format ostream ")~%")))

(defun change-names-to-pddl (problem domain)
  (sublis (case domain
	      ((logistics logistics-aips98-round2)
	       '((object . package)
		 (post-office . location)
		 (loc-at . in-city)
		 (inside . in)))
	    (blocksworld '((object . block)
			   (on-table . ontable)
			   (arm-empty . handempty)))
	    (t nil))
	  problem))
