(in-package "USER")

;; Example of use:
;; 1. (domain 'blocksworld)
;; 2. (problem 'suss)
;; 3. (run)
;; 4. (learn-macro 'blocksworld)
;;  It will generate a file macros.lisp in the blocksworld directory, that
;; you can insert into the domain file (before the first operator definition) if parameter install-macro-p is t

(defun learn-macro (domain
		    &optional install-macro-p printp
		    (first-step-number 1)
		    (last-step-number (length (prodigy-result-solution *prodigy-result*))))
  (let ((plan (visible-plan
	       (subseq (prodigy-result-solution *prodigy-result*)
		       (1- first-step-number) last-step-number)))
	(macros nil)
	(domain-directory (format nil "~(~a~a/~)" *world-path* domain)))
    (setf *macros-range* 1)
    (generate-new-op (create-macro plan)
		     (get-params-from-plan plan)
		     domain-directory
		     'prodigy
		     printp)
    (when install-macro-p
	(eval `(shell ,(format nil "cp -p ~adomain.lisp ~adomain-without-macros.lisp" domain-directory domain-directory)))
      (with-open-file (istream (format nil "~amacros.lisp" domain-directory)
				 :direction :input)
	  (do ((macro (read istream nil 'eof) (read istream nil 'eof))
	       (macros nil))
	      ((eq macro 'eof)
	       (with-open-file (istream2 (format nil "~adomain-without-macros.lisp" domain-directory)
					:direction :input)
		 (with-open-file (ostream (format nil "~adomain.lisp" domain-directory)
					  :direction :output
					  :if-exists :supersede
					  :if-does-not-exist :create)
		   (do ((domain-def (read istream2 nil 'eof) (read istream2 nil 'eof))
			(headerp t))
		       ((eq domain-def 'eof))
		     (if (and (listp domain-def)
			      (eq (car domain-def) 'operator))
			 (cond (headerp (setq headerp nil)
					(format ostream "~2%")
					(format ostream "~(~a~)" (car macros))
					(format ostream "~2%~(~s~)" domain-def))
			       (t (format ostream "~2%~(~s~)" domain-def)))
			 (format ostream "~2%~(~s~)" domain-def))))))
	    (push macro macros))))
    	(setf *valid-macros* 1)
    ))
		     

;;; *********************************************************************************
;;;               Functions for computing the visible form of a plan
;;; *********************************************************************************

(defun visible-plan (solution)
  (let ((plan nil)
	(preconds nil)
	(applied-node nil)
	(adds nil)
	(dels nil)
	(binding-node nil))
    (dolist (instantiated-op solution)
      (setq binding-node (p4::instantiated-op-binding-node-back-pointer instantiated-op))
      (setq preconds (preconds-from-instantiated-op binding-node))
      (setq applied-node (applied-node-from-instantiated-op binding-node instantiated-op))
      (setq adds (adds-from-applied-node applied-node))
      (setq dels (dels-from-applied-node applied-node))
      (push (list (get-applicable-op instantiated-op) preconds adds dels) plan))
    (reverse plan)))

(defun preconds-from-instantiated-op (binding-node)
  (get-binding-preconds (p4::binding-node-instantiated-preconds binding-node)))

(defun applied-node-from-instantiated-op (node instantiated-op)
  (if (and (p4::applied-op-node-p node)
	   (eq (p4::applied-op-node-instantiated-op node)
	       instantiated-op))
      node
      (let ((children (p4::nexus-children node)))
	(if children
	    (or (applied-node-from-instantiated-op (car children) instantiated-op)
		(applied-node-from-instantiated-op (cdr children) instantiated-op))))))

(defun adds-from-applied-node (node)
  (get-binding-preconds (cons 'and (p4::op-application-delta-adds (car (p4::applied-op-node-applied node))))))

(defun dels-from-applied-node (node)
  (get-binding-preconds (cons 'and (p4::op-application-delta-dels (car (p4::applied-op-node-applied node))))))


;;; **********************************************************************************************
;;;   Functions for creating a triangular table from a plan and creating a macroop from it
;;; **********************************************************************************************

(defun create-macro (plan)
  (let* ((num-ops (length plan))
	 (macro-array (make-array (list (1+ num-ops)
					(1+ num-ops))
				  :initial-element nil)))
    (do* ((i 0 (1+ i))
	  (solution plan (cdr solution))
	  (op (car solution)
	      (car solution))
	  (adds-from-previous (adds-from-previous macro-array i)
			      (adds-from-previous macro-array i))
	  (dels-from-previous (dels-from-previous macro-array i)
			      (dels-from-previous macro-array i)))
	((>= i num-ops) macro-array)
      (setf (aref macro-array (1+ i) i)
	    (op-name op))
      (setf (aref macro-array (1+ i) (1+ i))
	    (list (op-adds op) (op-dels op)))
      (setf (aref macro-array 0 i)
	    (set-difference (op-preconds op)
			    adds-from-previous
			    :test #'equal))
      (dotimes (j i)
	(setf (aref macro-array (1+ j) (1+ i))
	      (list (set-difference (car (aref macro-array (1+ j) i))
				    (op-dels op)
				    :test #'equal)
		    (set-difference (cadr (aref macro-array (1+ j) i))
				    (op-adds op)
				    :test #'equal)))))))

;; macro-array is a triangular table
;; params-info is a bindings list with vars and types info
;; style can be either prodigy or pddl
(defun generate-new-op (macro-array params domain-directory
				    &optional (style 'prodigy) (printp nil))
  (let ((preconds nil)
	(adds-dels nil)
	(adds nil)
	(dels nil)
	(ops nil)
	(ops-in-plan (1- (car (array-dimensions macro-array)))))
    (when printp
      (format t "~2%Triangular table:")
      (print-table macro-array))
    (dotimes (i ops-in-plan)
      (pushnew (aref macro-array (1+ i) i) ops :test #'equal)
      (dolist (precond (aref macro-array 0 i))
	(pushnew precond preconds :test #'equal))
      (setq adds-dels (aref macro-array (1+ i) ops-in-plan))
      (dolist (add (car adds-dels))
	  (pushnew add adds :test #'equal))
      (dolist (del (cadr adds-dels))
	  (pushnew del dels :test #'equal)))
    (make-operator domain-directory
		   :style style
		   :params params
		   :preconds preconds
		   :adds (set-difference adds preconds :test #'equal)
		   :dels dels
		   :ops (reverse ops)
		   :printp printp)))

(defun make-operator (domain-directory &key style params preconds adds dels ops printp)
  (let ((substitution (mapcar #'(lambda (param)
				  (cons (car param) 
					(intern (if (eq style 'prodigy)
						    (format nil "~:@(<~a-~a>~)" (car param) (cdr param))
						  (format nil "~:@(?~a-~a~)" (car param) (cdr param))))))
			      params)))
    (setq params (sublis substitution params))
    (setq preconds (sublis substitution preconds))
    (setq adds (sublis substitution adds))
    (setq dels (sublis substitution dels))
    (with-open-file (ostream (format nil "~(~amacros.lisp~)" domain-directory)
			     :direction :output
			     :if-exists :append
			     :if-does-not-exist :create)
      (cond ((eq style 'prodigy)
	     (format ostream "~2%(OPERATOR macro-op-~(~a~)" (gensym))
	     (format ostream "~%   (params")
	     (dolist (var (mapcar #'cdr (reverse substitution)))
		 (format ostream " ~(~a~)" var))
	     (format ostream ")~%   (preconds (")
	     (dolist (param params)
		 (format ostream "~%              (~(~a~) ~(~a~))" (car param) (cdr param)))
	     (format ostream ")~%      (and")
	     (dolist (precond preconds)
		 (format ostream "~%        ~(~a~)" precond))
	     (format ostream "))~%   (effects nil~%      (")
	     (dolist (add adds)
		 (format ostream "~%       (add ~(~a~))" add))
	     (dolist (del dels)
		 (format ostream "~%       (del ~(~a~))" del))
	     (format ostream ")))")
	     (if printp
		 (format t "~%It comes from: ~(~a~)" ops)))
	    (t 
	     (format ostream "~2%(operator macro-op-~(~a~)~%   :preconds (and ~(~a~))"
		     (gensym) preconds)
	     (format ostream "~%   :effects (and ")
	     (dolist (add adds)
		 (format ostream "~%              ~(~a~)" add))
	     (dolist (del dels)
		 (format ostream "~%              ~((not ~a)~)" del))
	     (format ostream "))~%")
	     (if printp
		 (format t "~%It comes from: ~(~a~)" ops)))))))

(defun op-name (op)
  (car op))

(defun op-preconds (op)
  (cadr op))

(defun op-adds (op)
  (caddr op))

(defun op-dels (op)
  (cadddr op))

(defun adds-from-previous (macro-array i)
  (let ((adds nil))
    (dotimes (j i)
      (dolist (add (car (aref macro-array (1+ j) i)))
	(pushnew add adds :test #'equal)))
    adds))

(defun dels-from-previous (macro-array i)
  (let ((dels nil))
    (dotimes (j i)
      (dolist (del (cadr (aref macro-array (1+ j) i)))
	(pushnew del dels :test #'equal)))
    dels))

(defun get-params-from-plan (plan)
  (let ((bindings nil))
    (dolist (instantiated-op plan)
      (dolist (obj (cdar instantiated-op))
	(if (not (assoc obj bindings))
	    (push (cons obj (get-object-type obj))
		  bindings))))
    bindings))
      

(defun print-table (table &optional (salida t))
  (terpri salida)
  (let ((dimensions (car (array-dimensions table))))
    (dotimes (j dimensions)
      (let ((fila-real (- dimensions j)))
	(dotimes (i (if (zerop j)
			fila-real
			(1+ fila-real)))
	  (format salida "~a " (aref table i (- fila-real 1))))
	(terpri salida)))))
