;;; File with code for probabilities computation.
;;; Output files: prob_rules.lisp (rules probabilities) and probabilities.lisp
;;; ("operator" probabilities).
;;;
;;; Use:
;;;  1. (progn (domain 'domain) (reset-domain-variables 'domain))
;;;  2. (hamlet-learn domain probsets :probabilities probabilities
;;;                                   :use-probabilities use-probabilities)
;;; where probabilities=nil: no rules probabilities learning, or
;;;       probabilities=rules: rules  probabilities learning.
;;; and use-probabilities is not used.
;;;  3. (hamlet-test domain probsets :probabilities probabilities
;;;                                  :use-probabilities use-probabilities)
;;; where probabilities=nil: no ops. probabilities learning, or
;;;       probabilities=operators: ops. probabilities learning.
;;; and use-probabilities=nil: do not use any probabilities,
;;;     use-probabilities=operators: use ops. probabilities,
;;;     use-probabilities=rules: use rules probabilities, or
;;;     use-probabilities=both: use ops. and rules probabilities.

(in-package "USER")

;;; Crea las listas en donde van a estar todas las probabilidades de
;;; operators, goals y subgoals.  El primer valor de *prob-operator* va a
;;; corresponder a los operadores, mientras que el segundo va a corresponder a
;;; las metas. En la tercera dimension, la posicion 0 corresponde con el
;;; numero de veces intentadas, y la 1 con el numero de veces ejecutadas con
;;; exito

(defun crea-lista (ops goals)
  (let ((operadores (length ops)) 
	(metas (length goals)))
    (setf *prob-operator*
	  (make-array (list operadores metas 2) :initial-element 0))
    (setf *prob-goals*
	  (make-array (list metas 2) :initial-element 0))
    (setf *prob-subgoal*
	  (make-array (list operadores metas 2) :initial-element 0))))


;;; Inicializa las tablas de probabilidades a cero y el fichero en donde se
;;; guardan las mismas.
;;; Tambien inicializa el fichero y las variables que se van a utilizar
;;; a la hora de la ejecucion de problemas con probabilidades de reglas
;;; de control

(defun init-probabilities (domain load-domain-p)
  (when load-domain-p
    (setf *domain-directory* (format nil "~a~(~a~)/" *world-path* domain))
    (domain domain))
  (setf *all-goals* (cdr (get-all-goals)))
  (setf *all-operators* (cdr (get-all-operators)))
  (crea-nuevas-reglas)
  (crea-lista *all-operators* *all-goals*)
  (write-probs))
  

;;; Inicializa todos los valores de las reglas, asi como el fichero en
;;; el que se guardan

(defun crea-nuevas-reglas nil
  (setf *number-rules-probs* 0)
  (setf *prob-rules* nil)
  (setf *number-operators-decision* 0)
  (setf *number-goals-decision* 0)
  (setf *number-bindings-decision* 0)
  (setf *number-subgoal-decision* 0)
  (write-rules))


;;; Procedimiento principal, en el cual, dependiendo de la variable
;;; *optimal-learning-p* se hara un aprendizaje recorriendo todo el arbol
;;; (valor nil) o solo los node-best-childs (valor t)

(defun create-probs (raiz anterior)
  (if *optimal-learning-p*
      (probs-optimal-learning raiz anterior)
    (opera-lista raiz anterior))
  (write-probs))


;;; Aprendizaje en el caso de que solo se aprenda de los node-best-childs

(defun probs-optimal-learning (raiz anterior)


;;; Recorre solo los best-child-info (secuencia de nodos que llega a la
;;; solucion del problema)

  (if (caar (node-best-child-info raiz))
      (probs-optimal-learning (caar (node-best-child-info raiz)) raiz))


;;; Crea las probabilidades de los operator-node

  (if (and (eq (node-type raiz) 'operator-node)
	   (not (eq (node-info raiz (node-type raiz)) 'prodigy4::*finish*)))
	(inc-prob-optimal-operator raiz anterior))


;;; Crea las probabilidades de los goal-node

  (if (and (eq (node-type raiz) 'goal-node)
	   (not (eq (car (get-target-goal raiz)) 'prodigy4::done)))
	(inc-prob-optimal-goals raiz anterior))

   
;;; Crea las probabilidades de los binding-node y applied-op-node

  (if (or(and (eq (node-type raiz) 'binding-node)
	      (not (eq (car (node-info raiz(node-type raiz))) 
		       'prodigy4::*finish*)))
	 (eq (node-type raiz) 'applied-op-node))
	(inc-prob-subgoal raiz)))


;;; Funcion que partiendo de un nodo incrementa la probabilidad del
;;; operador cuyo aprendizaje es success (incrementa la variable
;;; intentado y exito), e incrementa tambien en uno la variable
;;; intentado de todos los hermanos cuyo aprendizaje ha sido failure,
;;; siempre con respecto a la meta en la cual se esta trabajando

(defun inc-prob-optimal-operator (raiz anterior)
  (let* ((operador-elegido (node-info raiz (node-type raiz)))
	 (meta-elegida (car(get-target-goal anterior)))
	 (n-goal (position meta-elegida *all-goals*))
	 (n-op (position operador-elegido *all-operators*))
	 (lista (cdr (p4::nexus-children anterior))))
    
;;; Incremento la probabilidad de la variable intentado y exito del operador 
;;; cuyo aprendizaje es success

    (inc-prob-success *prob-operator* n-op n-goal)

;;; Si hay hermanos del operador cuyo aprendizaje es failure, incremento
;;; el valor de la variable intentado

    (dolist (aux lista)
	    (when (eq (node-learning-type aux) 'failure)
		  (setq operador-elegido (node-info aux (node-type aux)))
		  (setq n-op (position operador-elegido 
				       *all-operators*))
		  (inc-prob-failure *prob-operator* n-op n-goal)))))


;;; Al igual que en el procedimiento anterior, se incrementa el valor de
;;; la variable intentado y exito de la meta cuyo aprendizaje es success,
;;; y se incrementa la variable intentado de todas las metas hermanas cuyo
;;; aprendizahe es failure

(defun inc-prob-optimal-goals (raiz anterior)
  (let* ((meta-elegida (car (get-target-goal raiz)))
	 (n-goal (position meta-elegida *all-goals*))
	 (lista (cdr (p4::nexus-children anterior))))

;;; Se incrementa la probabilidad de la variable intentado y exito de la meta

    (inc-prob-success-goals n-goal)

;;; Por cada meta hermana a la anterior cuyo aprendizaje sea failure, se
;;; incrementa la variable intentado

    (dolist (aux lista)
      (when (and (eq (node-learning-type aux) 'failure)
		 (eq (node-type aux) 'goal-node))
	(setq meta-elegida (car(get-target-goal aux)))
	(setq n-goal (position meta-elegida *all-goals*))
	(inc-prob-failure-goals n-goal)))))

  
;;; Funcion que recorre todo el arbol en busca de todos los nodos del mismo.
;;; Se ejecuta cuando la variable *optimal-learning-p* tiene el valor nil.
;;; Recorre todo el arbol, en incrementa los valores de las variable intentado
;;; y exito de todos los nodos cuyo aprendizaje haya sido success.
;;; Cuando el aprendizaje haya sido failure, solo incrementa la variable
;;; intentado.

(defun opera-lista (raiz anterior)

;;; Recorre todo el arbol de busqueda expandido por PRODIGY

    (dolist (child (p4::nexus-children raiz))
	    (opera-lista child raiz))

;;; Crea las probabilidades de los operator-node

    (if (and (eq (node-type raiz) 'operator-node)
	     (not (eq (node-info raiz (node-type raiz)) 'prodigy4::*finish*)))
	(inc-prob-operator raiz (car(get-target-goal anterior))))

;;; Crea las probabilidades de los goal-node

    (if (and (eq (node-type raiz) 'goal-node)
	     (not (eq (car (get-target-goal raiz)) 'prodigy4::done)))
	(inc-prob-goals raiz))

;;; Crea las probabilidades de los binding-node y applied-op-node

    (if (or(and (eq (node-type raiz) 'binding-node)
		(not (eq (car (node-info raiz(node-type raiz))) 
			 'prodigy4::*finish*)))
	   (eq (node-type raiz) 'applied-op-node))
	  (inc-prob-subgoal raiz)))


;;; Incrementa la probabilidad de cada operador en la tabla *prob-operator*,
;;; dependiendo de si el tipo de aprendizaje es bueno (success) o el nodo es 
;;; un nodo de fallo (faillure)
;;; Si es sucess, se incrementa tando la variable exito como la de intentado. 
;;; En el caso contrario, solo se incrementa la variable intentado

(defun inc-prob-operator (raiz meta)
  (let* ((aprendizaje (node-learning-type raiz))
	(operador-elegido (node-info raiz (node-type raiz)))
	(number-goal (position meta *all-goals*))
	(number-op (position operador-elegido *all-operators*)))
   
;;; Dependiendo del aprendizaje se incrementa la variable intentado, o intentado
;;; y exito, de cada operador con respecto a la meta que esta intentando hacer
;;; cierta

    (case aprendizaje
	  (success
	   (inc-prob-success *prob-operator* number-op number-goal))
	  (failure
	   (inc-prob-failure *prob-operator* number-op number-goal)))))
	   

;;; Suma uno de las dos posiciones del array (intentado y exito), que se 
;;; pasa como parametro en la variable lista (tendria el nombre de la matriz
;;; de probabilidades a actualizar)

(defun inc-prob-success (lista op goal)
  (setf (aref lista op goal 0)
	(incf (aref lista op goal 0)))
  (setf (aref lista op goal 1)
	(incf (aref lista op goal 1))))


;;; Suma uno a la posicion del array de intentado, de la matriz de probabilidad
;;; dada en la variable lista

(defun inc-prob-failure (lista op goal)
    (setf (aref lista op goal 0)
	  (incf (aref lista op goal 0))))


;;; Incrementa la probabilidad de cada meta en la tabla de probabilidades
;;; *prob-goals*, dependiendo de si el tipo de aprendizaje es bueno (success) 
;;; o el un nodo de fallo (faillure).
;;; Si es sucess, se incrementa tando la variable exito como la de intentado. 
;;; En el caso contrario, solo se incrementa la variable intentado

(defun inc-prob-goals (raiz)
  (let ((aprendizaje (node-learning-type raiz))
	(number-goal (position (car (get-target-goal raiz)) 
			       *all-goals*)))
    
;;; Dependiendo del aprendizaje, se incrementa la variable intentado, o intentado
;;; y exito de la meta que se esta intentado hacer cierta

    (case aprendizaje
      (success
       (inc-prob-success-goals number-goal))
      (failure
       (inc-prob-failure-goals number-goal)))))


;;; Suma uno de las dos posiciones del array *prob-goals*, tanto a la de 
;;; intentados, como a la de exito, debido a que el tipo de aprendizaje ha
;;; sido success

(defun inc-prob-success-goals (goal)
  (setf (aref *prob-goals* goal 0)
	(incf (aref *prob-goals* goal 0)))
  (setf (aref *prob-goals* goal 1)
	(incf (aref *prob-goals* goal 1))))


;;; Suma uno a la posicion del array *prob-goals* de intentado, debido a que
;;; el tipo de aprendizaje ha sido failure

(defun inc-prob-failure-goals (goal)
  (setf (aref *prob-goals* goal 0)
	(incf (aref *prob-goals* goal 0))))


;;; Incrementa la probabilidad de la matriz *prob-subgoal*, qu es la que va a
;;; contener si es mejor hacer subgoal o aplicar el operador

(defun inc-prob-subgoal (raiz)
  (let ((aprendizaje nil)
	(operador-elegido nil)
	(meta-elegida nil)
	(number-goal nil)
	(number-op nil)
	(lista (p4::nexus-children raiz))
	(applied nil))

;;; Me aseguro de que en los hijos del nodo hay algun applied-op-node

    (dolist (aux lista)
      (if (and (eq (node-type aux) 'applied-op-node)(> (length lista) 1)) 
	  (setq applied aux)))
  
;;; Si hay algun applied-op-node incremento la probabilidad de failure o
;;; de success, segun el tipo de aprendizaje del applied-op-node

    (when applied 
      (setq aprendizaje (node-learning-type applied))
      (setq operador-elegido (node-info applied (node-type applied)))
      (setq number-op (position operador-elegido 
				*all-operators*))
      (case aprendizaje
	(success

;;; Incremente la probabilidad de fallo, ya que el applied era success
	     
	 (dolist (aux lista)
	   (when (not (eq (node-type aux) 'applied-op-node))
	     (setq meta-elegida (car (get-target-goal aux)))
	     (setq number-goal (position meta-elegida 
					 *all-goals*))
	     (inc-prob-failure *prob-subgoal* 
			       number-op number-goal))))

;;; Como el applied era failure ,si los goal-node son success, 
;;; se incrementa la probabilidad de success

	(failure
	 (dolist (aux lista)
	   (when (not (eq (node-type aux) 'applied-op-node))
	     (when (eq (node-learning-type aux) 'success)
	       (setq meta-elegida (car (get-target-goal aux)))
	       (setq number-goal (position meta-elegida 
					   *all-goals*))
	       (inc-prob-success *prob-subgoal* 
				 number-op number-goal)))))))))



;;; Devuelve la probabilidad del operador y la meta que se pasan como parametros
;;; con respecto al array de probabilidades que se le pasa en el parametro lista.
 
(defun get-prob (lista op goal)
  (if (and (not (zerop (aref lista op goal 0)))
	   (not (zerop (aref lista op goal 1))))
      (/ (aref lista op goal 1) 
	 (aref lista op goal 0))
      0))


;;; Devuelve la probabilidad de la meta que se pasa como parametro del array
;;; *prob-goals*

(defun get-prob-goals (goal)
  (if (and (not (zerop (aref *prob-goals* goal 0)))
	   (not (zerop (aref *prob-goals* goal 1))))
      (/ (aref *prob-goals* goal 1) 
	 (aref *prob-goals* goal 0))
      0))

;;; Devuelve el nombre de la meta que corresponde con el numero que se
;;; le ha pasado como parametro

(defun get-name-goal (num)
  (let ((list *all-goals*))
    (dotimes (< num)
	     (setq list (cdr list)))
    (car list)))


;;; Devuelve el nombre del operador que corresponde con el numero que se
;;; le ha pasado como parametro

(defun get-name-op (num)
  (let ((list *all-operators*))
    (dotimes (< num)
      (setq list (cdr list)))
    (car list)))


;;; Devuelve una lista con todos los operadores ordenados por el valor
;;; de la probabilidad del operador con respecto a la meta

(defun get-operators (node)
  (let ((n-ops (length *all-operators*))
	(n-goal (position (car (node-info node (node-type node))) 
			  *all-goals*))
	(cont 0) ;; Guarda el numero de valores diferentes de cero
	(lista nil)
	(salida nil)
	(aux 0)
;;;	(aux1 nil)
	(siguiente nil)
	(op nil)
	(meta nil)
	(a nil))

;;; Mira los operadores que tienen una probabilidad mayor que el umbral, y los 
;;; ordena con respecto a su probabilidad con respecto a la meta
    
    (dotimes (i n-ops)
	     (when (> (get-prob *prob-operator* aux n-goal) *umbral-prob*)
		   (setq lista (append lista 
				       (list (get-prob *prob-operator* 
						       aux n-goal))))
		   (incf cont))
	     (incf aux))
    (setq lista (sort lista #'>))
    
;;; Devuelve la lista con los nombres de los operadores seleccionados
    
    (setq salida nil)
    (setq siguiente lista)
    (dolist (aux lista)
	    (setq siguiente (cdr siguiente))
	    
;;; Principio para el caso de que sean iguales (misma probabilidad)
	    
	    (when (and (car siguiente)
		       (= aux (car siguiente)))
		  (setq meta 0)
		  (setq op 0)
		  (setq a 0)
		  (dotimes (< n-ops)
			   (when (= (get-prob *prob-operator* a n-goal) aux)
				 (when (and(> (aref *prob-operator* a n-goal 1) 
					      meta)
					   (eq (member (get-name-goal a) salida)
					       nil))
				       (setq meta (aref *prob-operator* a n-goal 1))
				       (setq op a)))
			   (incf a))
		  (setq salida (append salida (list (get-name-op op)))))
	    
;;; Caso en el que son diferentes
	    
	    (when (and (car siguiente)
		       (not (= aux (car siguiente))))
		  (setq a 0)
		  (dotimes (< n-ops)
			   (if (and (= (get-prob *prob-operator* a n-goal) aux)
				    (eq (member (get-name-op a)salida)nil))
			       (setq salida (append salida 
						    (list (get-name-op a)))))
			   (incf a))))
    
;;; Inserta la probabilidad menor de todas
    
    (when (not (car siguiente))
	  (setq a 0)
	  (setq meta 2)
	  (dotimes (< n-ops)
		   (if (and (< (get-prob *prob-operator* a n-goal) meta)
			    (> (get-prob *prob-operator* a n-goal) 0))
		       (setq meta (get-prob *prob-operator* a n-goal)))
		   (incf a))
	  (setq a 0)
	  (dotimes (< n-ops)
		   (if (and (= (get-prob *prob-operator* a n-goal) meta)
			    (eq (member (get-name-op a)salida)nil))
		       (setq salida (append salida (list (get-name-op a)))))
		   (incf a)))

;;; Devuelve los mejores operadores
    
    salida))


;;; Devuelve una lista con todos los operadores que tienen probabilidad de
;;; poderse ejecutar. Esta lista estara ordenada por la probabilidad
;;; de cada operador de poderse ejecutar, a mayor probabilidad de poderse
;;; ejecutar, estara en una posicion mas adelantada en la lista que devuelve

(defun best-operators (goal node ctrlrules)
  (let ((n-goal (position (car (node-info node (node-type node))) 
			  *all-goals*))
	(salida nil)
	(aux 0)
	(ops (p4::get-all-ops goal node)))
   
    
    (if (or (eq *use-probabilities* 'both) (eq *use-probabilities* 'operators))
	(setq salida (get-operators node))
      (setq salida (return-operators nil ops)))
     
;;; Introduzco solo como reglas de control las que annaden un operador posible
;;; dada la meta actual, y cumplen el lhs

    (when (and (or (eq *use-probabilities* 'rules) (eq *use-probabilities* 'both))
	       ctrlrules)
      (setq aux nil)
      (dolist (crt ctrlrules)
	(if (p4::match-lhs (p4::control-rule-if crt) node)
	    (setq aux (append aux (list crt)))))
      (setq ctrlrules aux)
	  
;;; Hasta ahora tenemos las reglas de control que se pueden ejecutar dada
;;; la meta actual
     

;;; Si hay reglas de control
     
      (when ctrlrules
	(setq ctrlrules (select-rules ctrlrules *number-operators-decision*))
	(setq salida (modifica-salida-operators salida ctrlrules n-goal
						*number-operators-decision*))))
		
	  
;;; Devuelve la lista con los mejores operadores

    (return-operators salida ops)))


;;; Devuelve la lista de salida con los mejores operadores 

(defun return-operators (name-operators ops)
  (let ((lista nil))
    (if name-operators
	(dolist (aux name-operators)
		(dolist (aux1 ops)
			(if (eq aux (p4::rule-name aux1))
			    (setq lista (append lista (list aux1))))))
      (dolist (aux ops)
	      (setq lista (append lista (list (p4::rule-name aux))))))
    
    lista))

;;; Identica a la funcion get-operators, pero con las metas
    
(defun get-goals (operator)
  (let ((n-goals (length *all-goals*))
	;; 	(n-ops nil)
	(cont 0) ;; Guarda el numero de valores diferentes de cero
	(lista nil)
	(salida nil)
	(siguiente nil)
	(meta nil)
	(a nil)
	(go nil)
	(aux 0)
	(goals (p4::give-me-all-pending-goals operator)))

;;; Compruebo que la probabilidad de las metas sea mayor que el umbral de
;;; probabilidad

    (dotimes (ng n-goals)
      (when (> (get-prob-goals aux) *umbral-prob*)
	(setq lista (append lista (list (get-prob-goals aux))))
	(incf cont))
      (incf aux))
    (sort lista #'>)

;;; Hasta aqui tenemos todas las probabilidades de metas ordenadas

    (setq salida nil)
    (setq siguiente lista)
    (dolist (aux lista)
      (setq siguiente (cdr siguiente))

;;; Principio para el caso de que sean iguales (misma probabilidad)

      (when (and (car siguiente)
		 (= aux (car siguiente)))
	(setq meta 0)
	(setq a 0)
	(dotimes (< n-goals) 
	  (when (= (get-prob-goals a) aux)
	    (when (and(> (aref *prob-goals* a 1) meta)
		      (eq (member (get-name-goal a) salida)nil))
	      (setq meta (aref *prob-goals* a 1))
	      (setq go a)))
	  (incf a))
	(setq salida (append salida (list (get-name-goal go)))))

;;; Caso en el que son diferentes

      (when (and (car siguiente)
		 (not (= aux (car siguiente))))
	(setq a 0)
	(dotimes (< n-goals)
	  (if (and (= (get-prob-goals a) aux)
		   (eq (member (get-name-goal a)salida)nil))
	      (setq salida (append salida (list (get-name-goal a)))))
	  (incf a))))

;;; Inserta la probabilidad menor de todas

    (when (not (car siguiente))
      (setq a 0)
      (setq meta 2)
      (dotimes (< n-goals)
	(if (and (< (get-prob-goals a) meta)
		 (> (get-prob-goals a) 0))
	    (setq meta (get-prob-goals a)))
	(incf a))
      (setq a 0)
      (dotimes (< n-goals)
	(if (and (= (get-prob-goals a) meta)
		 (eq (member (get-name-goal a)salida)nil))
	    (setq salida (append salida (list (get-name-goal a)))))
	(incf a)))

    (setq lista nil)
    (dolist (aux salida)
      (dolist (aux1 goals)
	(if (eq aux (p4::literal-name aux1))
	    (setq lista (append lista (list aux1))))))

;;; Devuelvo las metas ordenadas por su probabilidad

    lista))



;;; Devuelve una lista con todas las metas que se pueden ejecutar, ordenada
;;; por su probabilidad de poderse ejecutar, como en el procedimiento de
;;; best-operators.

(defun best-goals (operator ctrlrules)
  (let ((aux nil)
	(salida nil)
	(goals (p4::give-me-all-pending-goals operator)))

    (if (or (eq *use-probabilities* 'both) (eq *use-probabilities* 'operators))
	(setq salida (get-goals operator))
      (setq salida goals))
   
     
;;; Introduzco solo como reglas de control las que annaden un operador posible
;;; dada la meta actual, y cumplen el lhs

    (when (and (or (eq *use-probabilities* 'rules) (eq *use-probabilities* 'both))
	       ctrlrules)
	  (dolist (crt ctrlrules)
		  (if (p4::match-select-goal operator crt)
		      (setq aux (append aux (list crt)))))
	  (setq ctrlrules aux)
	  
;;; Hasta ahora tenemos las reglas de control que se pueden ejecutar dada
;;; la meta actual
     
  
;;; Si hay reglas de control
     
	  (when ctrlrules
		(setq ctrlrules (select-rules ctrlrules *number-goals-decision*))
		(setq salida (modifica-salida-goals salida ctrlrules *number-goals-decision*))))

;;; Devuelvo todas las metas ordenadas por su probabilidad
    salida))
		


;;; Devuelve si es mejor aplicar un operador o hacer subgoal en una meta,
;;; con respecto a las probabilidades que se encuentran en la matriz
;;; *prob-subgoal*. En caso de que se dieran los dos casos, devolveria
;;; apply o subgoal dependiendo de los valores de esta matriz de probabilidad,
;;; siempre que la la probabilidad de la matriz sea mayor que el valor de
;;; la variable *umbral-subgoal*

(defun select-apply-or-subgoal (ops goals rules) 
  (let ((operador-elegido (p4::rule-name (p4::instantiated-op-op (car ops))))
	(number-goal nil)
	(number-op nil)
	(salida ':apply)
	(reglas nil)
	(pr-rule 0)
	(pr-subgoal 0)
	(prob-rule-subgoal 0)
	(prob nil)
	(aux nil))

;;; Si no hay metas se devuelve apply

    (if (not goals)
       (setq salida ':apply))

    (when (or (eq *use-probabilities* 'operators) (eq *use-probabilities* 'both))
	  (when (and ops goals)
		(setq salida ':apply)
		(dolist (aux goals)
			(setq number-goal (position (p4::literal-name aux)
						    *all-goals*))
			(setq number-op (position operador-elegido
						  *all-operators*))
			(if (> (get-prob *prob-subgoal* number-op number-goal) 
			       *umbral-subgoal*)
			    (setq salida ':sub-goal)))))

    (when (or (eq *use-probabilities* 'rules) (eq *use-probabilities* 'both))
	  (when (and (or (eq *use-probabilities* 'rules) 
			 (eq *use-probabilities* 'both))
		     rules)
		(setq aux nil)

;;; Compruebo que el lhs de la regla es cierto en el estado actual

		(dolist (rule rules)
			(if (p4::match-lhs (p4::control-rule-if rule) nil)
			    (setq aux (append aux (list rule)))))
		(setq reglas aux)

		(when (and reglas goals ops)
		      (setq reglas (select-rules reglas *number-subgoal-decision*))

;;; Ordeno las reglas de control con respecto a la siguiente funcion:
;;;    Probabilidad = (*peso-reglas* * prob-regla) + (*peso-probs* * prob-apply-s)

		      (setq aux nil)
		      (dolist (re reglas)
			      (setq number-goal (position (p4::literal-name (car goals)) 
							  *all-goals*))
			      (setq number-op (position (p4::rule-name 
							 (p4::instantiated-op-op 
							  (car ops)))
							*all-operators*))
			      (setq pr-rule (get-prob-rules (get-number-rule 
							     (p4::control-rule-name 
							      (car reglas)))
							    *number-subgoal-decision*))
			      (setq pr-subgoal (get-prob *prob-subgoal* number-op number-goal))
			      (if (= pr-subgoal 0)
				  (setq prob-rule-subgoal (* *peso-reglas* pr-rule))
				(setq prob-rule-subgoal (+ (* *peso-reglas* pr-rule)
							   (* *peso-probs* pr-subgoal))))
			      (setq prob (append prob (list prob-rule-subgoal)))
			      (setq aux (append aux (list prob-rule-subgoal re))))
			      

;;; Devuelve el valor que marca la regla de control con mayor probabilidad

		      (sort prob #'>)
		      (setq reglas nil)
		      (when aux
			    (if (= (car prob) (car aux))
				(setq reglas (append reglas (list (cadr aux)))))
			    (setq aux (cddr aux)))

		      (if (eq (cadr (p4::control-rule-then (car reglas))) 
			      'sub-goal)
			  (setq salida ':sub-goal)
			(setq salida ':apply)))))
  
;;; Devuelve el resultado final (apply o subgoal)

    salida))


;;; Escribe las probabilidades de los tres arrays de probabilidad en un fichero,
;;; de nombre probabilities.lisp, que se encuentra en el directorio dado por la
;;; variable *domain-directory*
;;; Los dos primeros datos seran el nombre de todas las metas y de todos los
;;; operadores del dominio actual

(defun write-probs nil
  (with-open-file (impreso (format nil "~aprobabilities.lisp" 
				   *domain-directory*)
		   :direction :output
		   :if-exists :supersede
		   :if-does-not-exist :create)
		  
    (print *all-goals* impreso)
    (print *all-operators* impreso)

    (let ((operadores (length *all-operators*))
	  (metas (length *all-goals*)))
      (dotimes (op operadores)
	(dotimes (me metas)
	  (dotimes (num 2)
	    (print (aref *prob-operator* 
			 op me num) impreso))))
      (dotimes (me metas)
	(dotimes (num 2)
				      
	  (print (aref *prob-goals*
		       me num) impreso)))
      (dotimes (op operadores)
	(dotimes (me metas)
	  (dotimes (num 2)
	    (print (aref *prob-subgoal* 
			 op me num) impreso)))))))
 


;;; Recupera las probabilidades de los tres array de probabilidad desde el 
;;; fichero probabilities.lisp, y las introduce en los tres arrays de 
;;; probabilidades
;;; Recupera el valor de las variables globales que contienen las metas
;;; y operadores del dominio actual
;;; Tambien se leen las probabilidades de las reglas de control, y todas las
;;; variables globales que las reglas de control conllevan

(defun read-probs nil
  (with-open-file (impreso (format nil "~aprobabilities.lisp" 
				   *domain-directory*)
		   :direction :input)
		  
    (setf *all-goals* (read impreso))
    (setf *all-operators* (read impreso))		  
		  
    (crea-lista *all-operators* *all-goals*)
		  
    (let ((operadores (length *all-operators*))
	  (metas (length *all-goals*)))

      (format t "~%Reading probabilities of ~a operators and ~a goals~%"
	      operadores metas)

      (dotimes (op operadores)
	(dotimes (me metas)
	  (dotimes (num 2)
	    (setf (aref *prob-operator* 
			op me num)
	      (read impreso)))))
      (dotimes (me metas)
	(dotimes (num 2)
	  (setf (aref *prob-goals* me num)
	    (read impreso))))
      (dotimes (op operadores)
	(dotimes (me metas)
	  (dotimes (num 2)
	    (setf (aref *prob-subgoal* 
			op me num)
	      (read impreso)))))))
  
  (read-rules))
   
  

;;; *****************************************************************************
;;; ************ PRINCIPIO DE LA SEGUNDA PARTE (REGLAS DE CONTROL) **************
;;; *****************************************************************************



;;; Escribe las reglas aprendidas en el fichero prob_rules.lisp, asi como las
;;; variables globales que las reglas conllevan

(defun write-rules nil
  (with-open-file (impreso (format nil "~aprob_rules.lisp" 
				   *domain-directory*)
		   :direction :output
		   :if-exists :supersede
		   :if-does-not-exist :create)

;;; Guardo en el fichero el numero de reglas que vienen a continuacion

;;   (format t "~d ~d" *number-rules-probs* *number-operators-decision*)
		    (print *number-rules-probs* impreso)
		    (format t "~%Number of rules to be saved: ~a" *number-rules-probs* )
		    (dotimes (nrp *number-rules-probs*)
			     (dotimes (num 3)
				      (print (aref *prob-rules* nrp num) impreso)))

;;; Guardo los valores de *number-operators-decision*, .....

		    (print *number-operators-decision* impreso)
		    (print *number-goals-decision* impreso)
		    (print *number-bindings-decision* impreso)
		    (print *number-subgoal-decision* impreso)))


;;; Lee las reglas del fichero probs_rules.lisp, asi como las variables globales
;;; que van unidas a las reglas

(defun read-rules nil
  (let ((rules nil)
	(prob-file-name (format nil "~aprob_rules.lisp" *domain-directory*)))
    (cond ((probe-file prob-file-name)
	   (with-open-file (impreso prob-file-name :direction :input)
		    
;;; Leo el numero de reglas que me indica este valor del fichero
		    
	     (setq rules (read impreso))
	     (setf *number-rules-probs* rules)
	     (cond ((= rules 0)
		    (setf *prob-rules* nil))
		   ((> rules 0)
		    (setf *prob-rules* (make-array 
					(list *number-rules-probs* 3)))
		    (dotimes (ru rules)
		      (dotimes (num 3)
			(setf (aref *prob-rules* ru num)
			  (read impreso))))

;;; Leo las cuatro variables globales

		    (setf *number-operators-decision* (read impreso))
		    (setf *number-goals-decision* (read impreso))
		    (setf *number-bindings-decision* (read impreso))
		    (setf *number-subgoal-decision* (read impreso)))
		   (t nil))))
	  (t (setf *prob-rules* nil)))))



;;; Crea una nueva regla en la tabla de probabilidades de reglas (*prob-rules*),
;;; y actualiza la matriz de probabilidades


(defun nueva-regla-prob (name effects)
  (let* ((n-rules  *number-rules-probs*)
	 (aux (make-array (list (1+ n-rules) 3) :initial-element 1)))
  
    (incf *number-rules-probs*)
    (format t "~%Creating rule ~a~%" name)

;;; Inserta en la matriz de probabilidades (*prob-rules*) la nueva regla que
;;; se acaba de crear

    (when *prob-rules*

;;; Vuelca las antiguas reglas en la nueva variable

	  (dotimes (nr n-rules)
		   (dotimes (i 3)
			    (when (aref *prob-rules* nr i)
				  (setf (aref aux nr i) 
					(aref *prob-rules* nr i)))))

;;; Introduce la nueva regla

	  (setf (aref aux n-rules 1) name)

;;; La siguiente funcion hara lo que hacen los dos case. Hay que comprobar
;;; en las 3 funciones que llaman a nueva-regla-prob para que pasen el
;;; parametro effects del mismo tipo. Reemplazara a los dos case hasta el
;;; (setf *prob-rules* aux), esta funcion sin incluir (no se borra)

	  (insert-type-rule aux n-rules effects))
    
    (when (not *prob-rules*)
	  (setf *prob-rules* (make-array (list 1 3) :initial-element 1))
	  (setf (aref aux 0 1) name)

	  (insert-type-rule aux 0 effects))


;;; Esto se usa para incrementar las variables globales que guardan el numero
;;; total de veces que ha habido una regla de un determinado tipo. Seria 
;;; equivalente a todo lo de debajo

    (case (cadr effects)
	  (operators
	   (incf *number-operators-decision*))
	  (bindings
	   (incf *number-bindings-decision*))
	  (goals
	   (incf *number-goals-decision*))
	  (t
	   (incf *number-subgoal-decision*)))


;;; Incremento en uno el valor de una de las siguientes variables globales:
;;; *number-operators-decision*, *number-goals-decision*, 
;;; *number-bindings-decision*, *number-subgoal-decision*
;;; dependiendo del tipo de regla que sea

    (setf *prob-rules* aux)))


;;; Elimina la regla de probabilidad dada por la variable rule

(defun borra-regla-prob (rule node rule2)	
  (let* ((n-rules *number-rules-probs*)
	 (aux (make-array (list (1- n-rules) 3)))
	 (cont 0)
	 (old-rule nil)
	 (delete-rule (get-number-rule (p4::control-rule-name rule))))

   
;;; Comprueba que haya alguna regla en el nodo de decision a la cual se va
;;; a renombrar, y habria que sumar las probabilidades de una a la de otra

    (if node
	(if (car (decision-node-rules node))
	    (setq old-rule
		  (get-number-rule (p4::control-rule-name 
				    (car (decision-node-rules node)))))
	    (setq old-rule delete-rule))

	(setq old-rule (get-number-rule (p4::control-rule-name rule2))))

;;; Si entra a continuacion, implica que se borra la regla, pero no hay una regla
;;; a la cual sumar la probabilidad de la regla que se borra, con lo cual, habra
;;; que decrementar una de las variables globales de las que indican el numero 
;;; de veces que se ha creado una regla de un determinado tipo

    (when (eq old-rule delete-rule)
      (case (aref *prob-rules* old-rule 2)
	(operators
	 (decf *number-operators-decision* (aref *prob-rules* old-rule 0)))
	(bindings
	 (decf *number-bindings-decision* (aref *prob-rules* old-rule 0)))
	(goals
	 (decf *number-goals-decision* (aref *prob-rules* old-rule 0)))
	(t
	 (decf *number-subgoal-decision* (aref *prob-rules* old-rule 0)))))
	  
    
    (when delete-rule
      (format t "~%~%Deleting rule ~a  ~%" rule)

;;; Incremento las probabilidades de la regla que se queda con respecto a la
;;; que se borra (las que tienen diferente nombre pero hacen lo mismo)
	  
      (when (and (not (eq old-rule delete-rule)) old-rule)
	(setf (aref *prob-rules* old-rule 0)
	      (+ (aref *prob-rules* old-rule 0)
		 (aref *prob-rules* delete-rule 0))))

;;; Vuelca las antiguas reglas de conrol en la nueva lista
	  
      (dotimes (nr n-rules)
	(when (not (eq (get-name-rule nr)
		       (p4::control-rule-name rule)))
	  (setf (aref aux cont 0) (aref *prob-rules* nr 0))
	  (setf (aref aux cont 1) (aref *prob-rules* nr 1))
	  (setf (aref aux cont 2) (aref *prob-rules* nr 2))
	  (incf cont)))

      (decf *number-rules-probs*)
      (setf *prob-rules* aux))))
	 


;;; Inserta en la lista de reglas el tipo de regla que es (bindings, ops,
;;; goals...). Hay que comprobar que las tres funciones que llaman a
;;; nueva-regla-prob pasen  bien el parametro para que el valor deseado sea el
;;; del cadr

(defun insert-type-rule (aux n-rules effects)
  (case (cadr effects)
	(operators
	 (setf (aref aux n-rules  2) 'operators))
	(bindings
	 (setf (aref aux n-rules 2) 'bindings))
	(goals
	 (setf (aref aux n-rules 2) 'goals))
	(t
	 (setf (aref aux n-rules 2) 'subgoal))))



;;; Devuelve la probabilidad de la regla dada por el parametro num

(defun get-prob-rules (num number-decision)
  (if (and (not (zerop (aref *prob-rules* num 0)))
	   (not (zerop number-decision)))
      (/ (aref *prob-rules* num 0) number-decision)
      0))


;;; Devuelve el nombre de la regla dada por el parametro num

(defun get-name-rule (num)
  (aref *prob-rules* num 1))


;;; Devuelve el numero de correlacion en la tabla de probabilidades, en el 
;;; cual esta la regla que se le pasa como parametro

(defun get-number-rule (rule)
  (let ((nrule *number-rules-probs*)
	(salida nil))
    (dotimes (nr nrule)
	     (if (not salida)
		 (if (eq (aref *prob-rules* nr 1) rule)
		     (setq salida nr))))
    salida))


;;; Devuelve una lista ordenada por el valor de las mejores probabilidades de
;;; todas las reglas de control que se le pasan en el parametro crtlrules

(defun select-rules (ctrlrules number-decision)
  (let ((probs nil)
	(siguiente nil)
	(meta nil)
	(lista nil)
	(rule nil)
	(salida nil)
	(n-ctrlrules *number-rules-probs*))
    (dolist (cr ctrlrules)
      (dotimes (ncr n-ctrlrules)
	(if (and (> (get-prob-rules ncr number-decision) *umbral-prob*)
		 (eq (get-name-rule ncr)(p4::control-rule-name cr)))
	    (setq probs
		  (append probs 
			  (list (get-prob-rules ncr number-decision)))))))
    
;;; Ordeno las reglas de control de mayor a menor

    (sort probs #'>)
    
    (setq siguiente probs)
    (dolist (aux probs)
      (setq siguiente (cdr siguiente))

;;; Principio para el caso de que sean iguales

      (when (and (car siguiente)
		 (= aux (car siguiente)))
	(setq meta 0)
	(setq rule 0)
	(dotimes (nr n-ctrlrules)
	  (when (= (get-prob-rules nr number-decision) aux)
	    (when (and (> (aref *prob-rules* nr 0) meta)
		       (incluido ctrlrules (get-name-rule nr))
		       (not (miembro salida (get-name-rule nr))))
	      (setq meta (aref *prob-rules* nr 0))
	      (setq rule nr))))
	(setq salida (append salida (list (get-name-rule rule)))))
       
;;; Principio para el caso de que son diferentes
    
      (when (and (car siguiente)
		 (not (= aux (car siguiente))))
	(dotimes (nr n-ctrlrules)
	  (if (and (= (get-prob-rules nr number-decision) aux)
		   (incluido ctrlrules (get-name-rule nr))
		   (not (miembro salida (get-name-rule nr))))
	      (setq salida (append salida 
				   (list (get-name-rule nr)))))))
	    
;;; Inserta la probabilidad menor de todas
		  
      (when (not (car siguiente))
	(dotimes (nr n-ctrlrules)
	  (if(and (= (get-prob-rules nr number-decision) aux)
		  (incluido ctrlrules (get-name-rule nr))
		  (not (miembro salida (get-name-rule nr))))
	     (setq salida (append salida 
				  (list (get-name-rule nr))))))))

        
;;; Devuelve la lista de salida con las mejores reglas de control 
    
    (setq lista nil)
    (dolist (aux salida)
      (dolist (aux1 ctrlrules)
	(if (eq aux (p4::control-rule-name aux1))
	    (setq lista (append lista (list aux1))))))
    
    lista))


;;; Devuelve t si la regla esta incluida en el conjunto de reglas de control, y nil
;;; en el caso contrario

(defun incluido (list rule)
  (let ((find nil))
    (dolist (aux list)
	    (if (and (not find)
		     (eq (p4::control-rule-name aux) rule))
		(setq find t)))
    find))


;;; Devuelve t si un elemento es mienbro de una lista, nil en el caso contrario

(defun miembro (list rule)
  (let ((find nil))
    (dolist (aux list)
	    (if (and (not find)
		     (eq aux rule))
		(setq find t)))
    find))
		     

;;; Devuelve el mejor operador a ejecutar, teniendo en cuenta las reglas de
;;; control  en el caso de que las haya

(defun modifica-salida-operators (lista rules n-goal number-op-decision)
  (let ((aux nil)
	(op-rule nil)
	(prob-op nil)
	(li nil)
	(prob nil)
	(number-op 0)
	(prob-rule-op 0)
	(pr-rule 0)
	(number-rules 0))
    
;;; Creo las nuevas probabilidades de reglas y operadores con respecto a la formula
;;;    Probabilidad = (*peso-reglas* * prob-regla) + (*peso-probs* * prob-op)

    (when (not (eq *use-probabilities* 'rules))
	  (dolist (ru rules)
		  (incf number-rules)
		  (setq number-op (position (cadddr (p4::control-rule-then ru))
					    *all-operators*))
		  (setq prob-op (get-prob *prob-operator* number-op n-goal))
		  (setq pr-rule (get-prob-rules (get-number-rule (p4::control-rule-name ru))
						number-op-decision))
		  (if (= prob-op 0)
		      (setq prob-rule-op (* *peso-reglas* pr-rule))
		    (setq prob-rule-op (+ (* *peso-reglas* pr-rule) (* *peso-probs* prob-op))))
		  (setq prob (append prob (list prob-rule-op)))
		  (setq aux (append aux (list prob-rule-op ru))))

;;; Hasta aqui tenemos todas las reglas de control, teniendo en cuenta tambien las
;;; probabilidades de los operadores. En la lista prob estan todas las probabilidades
;;; Ahora ordeno las reglas con respecto a estas probabilidades	  
	  
	  (sort prob #'>)
	  (setq rules nil)
	  (dolist (pr prob)
		  (setq li aux)
		  (dotimes (i number-rules)
			   (if (= pr (car li))
			       (setq rules (append rules (list (cadr li)))))
			   (setq li (cddr li)))))

;;; Hasta ahora tendriamos las reglas de control ordenadas segun la funcion anterior

;;; Modifica los operadores segun el orden que indica las reglas de control

    (setq aux nil)
    (dolist (ru rules)
	    (setq op-rule (cadddr (p4::control-rule-then ru)))
	    (dolist (li lista)
		    (if (and (eq li op-rule)
			     (not (include-list-op aux op-rule)))
			(setq aux (append aux (list li))))))
    
;;; Devuelvo la nueva lista

    aux))


;;; Es identica a la anterior, pero en vez de modificar operadores, se modifican
;;; metas

(defun modifica-salida-goals (lista rules number-goal-decision)
  (let ((aux nil)
	(goal-rule nil)
	(number-rules 0)
	(name-goals nil)
	(number-goal 0)
	(prob nil)
	(prob-goal 0)
	(prob-rule 0)
	(li nil)
	(prob-rule-goal 0))
   
;;; Crea las nuevas probabilidades de reglas y operadores con respecto a la formula
;;;      Probabilidad = (*peso-reglas* * prob-regla) + (*peso-probs* * prob-meta)

    (when (not (eq *use-probabilities* 'rules))
      (dolist (ru rules)
	(incf number-rules)
	(setq name-goals (caar (cdddr (p4::control-rule-then ru))))
	(setq number-goal (position name-goals *all-goals*))
	(setq prob-goal (get-prob-goals number-goal))
	(setq prob-rule (get-prob-rules (get-number-rule (p4::control-rule-name ru))
					number-goal-decision))
	(if (= prob-goal 0)
	    (setq prob-rule-goal (* *peso-reglas* prob-rule))
	  (setq prob-rule-goal (+ (* *peso-reglas* prob-rule) (* *peso-probs* prob-goal))))
	(setq prob (append prob (list prob-rule-goal)))
	(setq aux (append aux (list prob-rule-goal ru))))

;;; Hasta aqui tenemos todas las reglas de control, teniendo en cuenta tambien las 
;;; probabilidades de las metas

      (sort prob #'>)
      (setq rules nil)
      (dolist (pr prob)
	(setq li aux)
	(dotimes (i number-rules)
	  (if (= pr (car li))
	      (setq rules (append rules (list (cadr li)))))
	  (setq li (cddr li)))))

;;; Modifico las metas segun el orden que me indican las reglas de control

    (setq aux nil)
    (dolist (ru rules)
      (setq goal-rule (car (cadddr (p4::control-rule-then ru))))
      (dolist (li lista)
	(if (and (eq (p4::literal-name li) goal-rule)
		 (not (include-list-goals aux goal-rule)))
	    (setq aux (append aux (list li))))))
    
;;; Devuelvo la nueva lista

    aux))
	
   
;;; Devuelve t si el operador esta incluido en lalista

(defun include-list-op (list find)
  (let ((aux nil))
    (dolist (li list)
	    (if (eq li find)
		(setq aux t)))
    aux))


;;; Devuelve t si una meta esta incluida en la lista

(defun include-list-goals (list find)
  (let ((aux nil))
    (dolist (li list)
      (if (eq (prodigy4::literal-name li) find)
	  (setq aux t)))
    aux))




;;; **************************************************************
;;;                      UTILIDADES VARIAS
;;; **************************************************************


;;; Devuelve una tabla con todas las probabilidades de operadores y metas del
;;; array *prob-operator*

(defun probs-ops (goals)
  (let ((op nil)
	(go nil)
	(list-ops *all-operators*))

    (if (cddr goals)
	(probs-ops (cdr goals)))
    
    (dolist (aux list-ops)
      (setq op (position aux *all-operators*))
      (setq go (position (car goals) *all-goals*))
      (if (and (not (zerop (aref *prob-operator* op go 0))) 
	       (not (zerop (aref *prob-operator* op go 1)))) 
	  (format t "~%Goal: ~10a Operator: ~10a Probability: ~f" 
		  (car goals) aux
		  (/ (aref *prob-operator* op go 1) (aref *prob-operator* op go 0)))
	  (format t "~%Goal: ~10a Operator: ~10a Probability: 0" 
		  (car goals) aux)))))


;;; Devuelve las probabilidades de las metas, que se encuentran en la tabla
;;; de probabilidades llamada *prob-goals*

(defun probs-goals nil
  (let ((n-goals (length *all-goals*))
	(goal nil))
    (format t "~%")
    (dotimes (aux n-goals)
	     (setq goal (get-name-goal aux))
	     (if (and (not (zerop (aref *prob-goals* aux 1)))
		      (not (zerop (aref *prob-goals* aux 0))))
		 (format t "~%Goal: ~10a Probability: ~f" goal
			 (/ (aref *prob-goals* aux 1) (aref *prob-goals* aux 0)))
	       (format t "~%Goal: ~10a Probability: 0" goal)))))

;;; Devuelve una tabla con todas las probabilidades de operadores y metas
;;; del array *prob-operator*

(defun saca-prob nil
  (probs-ops *all-goals*)
  (probs-goals))