;;; This file has the code for automatically generating problems that are
;;; suitable for learning with Hamlet and EvoCK.

(in-package "USER")

;; Interface: generate-random-problems
;; Fitness function: valid-problem-p
;; Interface function with domain dependent random generators:
;;   generate-problem. It has to return t if it could generate a problem with
;;   the required characteristics, or nil if it could not. The problem will be
;;   the only one in the probset called "temp-problems.lisp" in the probsets
;;   directory of the domain.

;; Hamlet has to be loaded before.

;; probset: probset name where it will leave all generated problems
;; time-bound: of execution of problems by Prodigy
;; no-goals: number of goals or 'random
;; max-goals: if no-goals=random, then this is the maximum number of goals
;; types-info should be a list of sublists. Each sublist must be:
;;   (type no-objects-of-type-or-random max-no-objects-of-type-if-random
;;         min-no-objects increment-no-objects-for-difficulty)
;; parameters-alist: alist of parameters-values for each generator
;; output-format:
;;   - hamlet: each problem description is:
;;           (problem-name best-cost best-length problem)
;;   - evock: each problem description is just the problem
;; filter-problems-p: if we want to filter out problems that are too simple/complex to learn from
;; probset: if given it should be a list of problem definitions

;; Examples:
;; (generate-random-problems 'blocksworld 200 "prueba" 100 'random 2 '((object random 5 1 1)) '((max-tower-size . 5) (goal-as-complete-state-p . t) (type . classic)) 'hamlet)
;;    (generate-single-problem 'logistics 'random 50 5
;;                             '((object random 6) (city random 5)))
;; External-generator-p is t when one will use a PDDL random generator for instance.
(defun generate-random-problems (domain no-probs probset-name time-bound
				 no-goals max-goals types-info
				 parameters-alist output-format
				 &optional (filter-problems-p t)
					   probset
					   external-generator-p)
  (reset-domain-variables nil domain)
  (setf *multiple-solutions-flg* :shorter)
  (let ((temp-file (concatenate 'string *domain-directory*
				"probsets/temp-problems.lisp"))
	(problems-list-p (and probset t)))
    (with-open-file (ofile (concatenate 'string *domain-directory*
					"probsets/" probset-name "."
					*lisp-extension*)
		     :direction :output 
		     :if-exists :supersede
		     :if-does-not-exist :create)
      (format ofile ";;; Problem set of the ~a domain" domain)
      (format ofile "~%;;; Number of problems: ~a" no-probs)
      (format ofile "~%;;; Time bound: ~a" time-bound)
      (format ofile "~%;;; Number of goals: ~a" no-goals)
      (format ofile "~%;;; Max. number of goals: ~a" max-goals)
      (format ofile "~%;;; Types information: ")
      (dolist (type-info types-info)
	      (format ofile "~%;; ~a" type-info))
      (format ofile "~%;;; Parameters information: ~a" parameters-alist)
      (format ofile "~%;;; Filter problems: ~a" filter-problems-p)
      (format ofile "~2%(setf *all-problems* '(")

      (domain domain)
      (if (not problems-list-p)
	  (load (concatenate 'string *domain-directory* "gen-probs")))
      (do ((i 1)
	   (number-probs-generated 0 (1+ number-probs-generated))
	   (problem-generated-p nil))
	  ((or (= i (1+ no-probs))
	       (and problems-list-p
		    (not probset)))
	   (format ofile "~%))")
	   (format ofile "~2%;;; Number of generated problems: ~d"
		   number-probs-generated)
	   (format ofile "~%;;; Number of valid problems: ~d" no-probs))
	(setq problem-generated-p
	  (cond (problems-list-p
		 (with-open-file (temp-file-stream temp-file
				  :direction :output
				  :if-exists :supersede
				  :if-does-not-exist :create)
		   (format temp-file-stream
			   "(setf *all-problems* '(~a))"
			   (car probset)))
		 (setq probset (cdr probset))
		 t)
		(t (generate-problem no-goals max-goals types-info
				     "temp-problems" parameters-alist))))
	(if problem-generated-p
	    (cond (external-generator-p
		   (setq i (1+ i)))
		  (t 
		   (load temp-file)
		   (let* ((problem (car *all-problems*))
			  (problem-name (problem-name problem))
			  (new-problem-name nil)
			  (solution nil))
		     (when (or (not filter-problems-p)
			       (test-problem problem time-bound t 0 nil nil nil nil
					     *max-depth* nil nil 'lazy-partial))
		       (if filter-problems-p
			   (setq solution (prodigy-result-solution *prodigy-result*)))
		       (setq new-problem-name
			 (intern (format nil "~:@(~a-~d~)" problem-name i)))
		       (setf (problem-name problem) new-problem-name)
		       (if (and filter-problems-p (eq output-format 'hamlet))
			   (format ofile "~%~a"
				   (list new-problem-name *best-cost*
					 (if (listp solution)
					     (length solution)
					   1)
					 problem))
			 (format ofile "~%~a" problem))
		       (setq i (1+ i)))
		     (delete-file temp-file)))))))))

;; It will be defined in the corresponding domain files.
(defun generate-problem (no-goals max-goals types-info probset &optional parameters-alist)
  (declare (ignore no-goals max-goals types-info probset parameters-alist)))

;; If cr-file specified, it will use it
(defun test-problem (problem time-bound multiple-sols-p my-output-level
		     print-alts-p linearp chronological-back-p
		     random-behavior-p depth-bound cost-bound cost-type eager-mode
		     &optional cr-file)
  (declare (special *real-t-tree* *prodigy-result* *learnp* *my-current-node*
		    *stop-during-problem-generation-p*))
  (my-load-problem problem)
  (when cr-file
      (load cr-file)
    (load-domain))
  (my-run time-bound multiple-sols-p my-output-level print-alts-p linearp chronological-back-p
	  random-behavior-p depth-bound cost-bound cost-type)
  (let* ((time (prodigy-result-time *prodigy-result*))
	 (nodes (prodigy-result-nodes *prodigy-result*))
	 (solutionp (solutionp))
	 (solution (prodigy-result-solution *prodigy-result*))
	 (solutions (prodigy-result-solutions *prodigy-result*))
	 (num-ops (if (listp solution)
		      (length (prodigy-result-solution *prodigy-result*))
		    1))
	 (nodes-solution-path (if (and (numberp num-ops)
				       (> num-ops 0))
				  (+ 3 (* num-ops 4))
				0))
	 (whole-tree-expanded-p (whole-tree-expanded-p))
	 (learnp nil)
	 (valid-problem-p nil))
    (setf *real-t-tree* (p4::find-node 2))
    (setf (prodigy-result-nodes-solution-path) nodes-solution-path)
    (setq learnp *learnp*)
    (setf *learnp* nil)
    ;; so that it labels the tree for cheking some things in valid-problem-p
    (setf *my-current-node* nil)
    (learn-from-tree *real-t-tree* eager-mode whole-tree-expanded-p)
    (setf *learnp* learnp)
    (setq valid-problem-p
      (valid-problem-p time nodes solutionp num-ops nodes-solution-path
		       whole-tree-expanded-p solutions eager-mode))
    (if *stop-during-problem-generation-p*
	(break))
    valid-problem-p))

;; This function has to be parameterized to the type of problems that one
;; would like to have.  Function for filtering for EvoCK/Hamlet Now, it checks
;; whether there was a solution in that time bound, it expanded the whole
;; search tree, and the first path followed did not end in the best
;; solution (there was at least a backtrack)
(defun valid-problem-p (time nodes solutionp num-ops nodes-solution-path
			     whole-tree-expanded-p solutions eager-mode)
  (declare (ignore num-ops solutions time nodes nodes-solution-path))
  (and solutionp
       (or (eq eager-mode 'eager-partial)
	   (eq eager-mode 'lazy-partial)
	   whole-tree-expanded-p)
       (or (eq eager-mode 'eager-partial)
	   (eq eager-mode 'eager-total)
	   (longer-first-path-p))))

;; the number of operators in the solution was greater than 1, there were at
;; least two different-quality solutions (now, only solution length)
;; (defun valid-problem-p (time nodes solutionp num-ops nodes-solution-path
;; 			     whole-tree-expanded-p solutions)
;;   (declare (ignore time nodes-solution-path nodes solutions))
;;   (and solutionp whole-tree-expanded-p
;;        (two-different-quality-solutions-p solutions)
;;        (> num-ops 1)
;;        (longer-first-path-p)))

;; (> nodes (* 2 nodes-solution-path))


(defun two-different-quality-solutions-p (solutions)
  (not (apply #'= (mapcar #'(lambda (solution)
			      (if (listp solution)
				  (length solution)
				1))
			  solutions))))

(defun generate-random-value (info max-value)
  (if info
      (if (eq (car info) 'random)
	  (1+ (random max-value))
	  (car info))
      max-value))

;; Returns a similar problem to the previous one
;; For now, *domain-predicates* should refer to primitive types
;; We can change either the state or the goal. Probability of 0.6 of changing the state
(defun generate-problem-from-problem (problem)
  (let* ((type-to-change (choose-one *changeable-types*))
	 (objects (cdr (problem-objects problem)))
	 (instances-to-change (problem-objects-of-type type-to-change objects))
	 (instance-to-change (choose-one instances-to-change))
	 (statep (<= (random 1.0) *al-change-state-probability*))
	 (state (cdadr (problem-state problem)))
	 (goals (cadr (problem-goals problem)))
	 (literals (if statep
		       state
		     (if (eq (car goals) 'and)
			 (cdr goals)
		       (list goals))))
	 (type-finite-state (cdr (assoc type-to-change *finite-states*)))
	 (current-literal (find-instance-in-state instance-to-change literals type-finite-state))
	 (new-predicate (choose-one type-finite-state))
	 (predicate-info (car (member new-predicate *domain-predicates* :key #'car)))
	 (new-literals (cons 'and
			     (substitute (generate-new-literal new-predicate instance-to-change
							       objects
							       current-literal predicate-info nil)
					 current-literal literals :test #'equal))))
    (if statep
	(setf (problem-state problem) (list 'state new-literals))
      (setf (problem-goals problem) (list 'goal new-literals)))
    (with-open-file (ofile (concatenate 'string *domain-directory* "probsets/temp-problem.lisp")
		     :direction :output :if-exists :supersede :if-does-not-exist :create)
      (format ofile "~a" problem))))

;;     (values type-to-change
;; 	    instance-to-change
;; 	    current-literal
;; 	    new-predicate
;; 	    predicate-info
;; 	    new-literal)

(defun find-instance-in-state (instance state type-finite-state)
  (find-if #'(lambda (literal)
	       (and (member (car literal) type-finite-state)
		    (member instance (cdr literal))))
	   state))

(defun all-objects-of-type (types objects)
  (let ((result nil))
    (dolist (type (if (atom types) (list types) types))
      (dolist (object (problem-objects-of-type type objects))
	(push object result)))
    result))

(defun generate-new-literal (new-predicate instance-to-change objects
			     current-literal predicate-info use-frequencies-p)
  (cons new-predicate
	(mapcar #'(lambda (arg-literal arg-predicate)
		    (if (eq arg-literal instance-to-change)
			instance-to-change
		       (if use-frequencies-p
			   (choose-one-with-frequencies (remove arg-literal (all-objects-of-type arg-predicate objects))
							'instance)
			 (choose-diff-one arg-literal (all-objects-of-type arg-predicate objects)))))
		(cdr current-literal)
		(cdr predicate-info))))

;; Returns a similar problem to the previous one, but having in mind the
;; frecuencies of literals in the regressed-states, predicates and instances
;; that appear on those regressed-states
;; For now, *domain-predicates* should refer to primitive types
;; We can change either the state or the goal.
(defun generate-problem-from-rules (problem)
  (let* ((type-to-change (choose-one-with-frequencies *changeable-types* 'type))
	 (objects (cdr (problem-objects problem)))
	 (instances-to-change (problem-objects-of-type type-to-change objects))
	 (instance-to-change (choose-one-with-frequencies instances-to-change type-to-change))
	 (statep (<= (random 1.0) *al-change-state-probability*))
	 (state (cdadr (problem-state problem)))
	 (goals (cadr (problem-goals problem)))
	 (literals (if statep
		       state
		     (if (eq (car goals) 'and)
			 (cdr goals)
		       (list goals))))
	 (type-finite-state (cdr (assoc type-to-change *finite-states*)))
	 (current-literal (find-instance-in-state instance-to-change literals type-finite-state))
	 (new-predicate (choose-one-with-frequencies type-finite-state 'predicate))
	 (predicate-info (car (member new-predicate *domain-predicates* :key #'car)))
	 (new-literals (cons 'and
			     (substitute (generate-new-literal new-predicate instance-to-change
							       objects
							       current-literal predicate-info t)
					 current-literal literals :test #'equal))))
    (if statep
	(setf (problem-state problem) (list 'state new-literals))
      (setf (problem-goals problem) (list 'goal new-literals)))
    (with-open-file (ofile (concatenate 'string *domain-directory* "probsets/temp-problem.lisp")
		     :direction :output :if-exists :supersede :if-does-not-exist :create)
      (format ofile "~a" problem))))

;; It does a roulette over the elements of the list with respect to their
;; frequencies in the type
(defun choose-one-with-frequencies (choosing-list type)
  (declare (special *active-learning-hash*))
  (do* ((type-frequency (gethash type *active-learning-hash*))
	(random-number (random (1+ type-frequency)))
	(list choosing-list (cdr list))
	(element (car list) (car list))
	(element-frequency (gethash (cons type element) *active-learning-hash*)
			   (gethash (cons type element) *active-learning-hash*))
	(accumulated-frequencies element-frequency
				 (+ accumulated-frequencies element-frequency)))
      ((or (null list)
	   (<= random-number accumulated-frequencies))
       element)
    (format t "~%random: ~d; element: ~d; frequency: ~d; acc-freq: ~d"
	    random-number element element-frequency accumulated-frequencies)))
