;;;; Code to translate Prodigy problem descriptions into PDDL

(in-package "USER")

;;; Translates problem descriptions from a probset file in Prodigy4.0 into
;;; PDDL format
(defun translate-problems (file &key (domain 'logistics)
				     (output-directory *ff-output-directory*))
  (declare (special *all-problems*))
  (load file)
  (dolist (problem *all-problems*)
    (translate-one-problem problem nil domain output-directory)))

;; It translates an individual problem from a single problem file
(defun translate-problem (problem &key (domain 'logistics)
				       (output-directory *ff-output-directory*))
  (cond ((or (stringp problem) (pathnamep problem))
	 (if (probe-file problem)
	     (with-open-file (istream problem :direction :input)
	       (translate-one-problem (read istream) (pathname-name problem)
				      domain output-directory))
	   (format t "~% File not found: ~a" problem)))
	((listp problem)
	 (translate-one-problem problem (cadadr problem)
				domain output-directory))
	(t (format t "~%Nothing to translate ~a" problem)
	   nil)))

(defun translate-one-problem (problem name domain output-directory)
  (let* ((problem-description (caddr problem))
	 (problem-name (cadadr problem-description))
	 (objects-defs (cdaddr problem-description))
	 (state (nth 3 problem-description))
	 (goals (nth 4 problem-description))
	 (output-file (format nil "~(~a-~a~).pddl" (or name problem-name) domain)))
    (with-open-file (ostream (format nil "~(~a~a~)" output-directory output-file)
		     :direction :output :if-exists :supersede
		     :if-does-not-exist :create)
      (format ostream "(define (problem ~(~a~)) (:domain ~(~a~))~%" problem-name domain)
      (write-pddl-objects objects-defs domain ostream)
      (write-pddl-state state domain ostream)
      (write-pddl-goals goals domain ostream)
      (format ostream ")" problem-name domain))
    (values output-directory output-file)))

(defun write-pddl-objects (objects-defs domain ostream)
  (format ostream "  (:objects")
  (dolist (types objects-defs)
    (format ostream "~%    ")
    (let ((type (car (last types)))
	  (objects (butlast (if (eq (car types) 'objects-are)
				(cdr types)
			      types))))
      (dolist (object objects)
	(format ostream " ~(~a~)" object))
      (format ostream " - ~a" type)))
  (format ostream "  )~%"))

;;  (format ostream " - ~a"(change-names type domain))

(defun write-pddl-state (state domain ostream)
  (format ostream "  (:init")
  (dolist (literal (cdadr state))
    (format ostream "~%    ~(~a~)" literal))
  (format ostream ")~%"))

;;   (dolist (literal (change-names (cdadr state) domain))

(defun write-pddl-goals (goals domain ostream)
  (let* ((the-goals (cadr goals))
	 (more-than-one-goal-p  (eq (car the-goals) 'and)))
    (format ostream "  (:goal")
    (if more-than-one-goal-p
	(format ostream "    (and"))
    (dolist (literal (if more-than-one-goal-p
			 (cdr the-goals)
			 (cdr goals)))
      (format ostream "~%      ~(~a~)" literal))
    (if more-than-one-goal-p
	(format ostream ")"))
    (format ostream ")~%")))

;;     (dolist (literal (change-names (if more-than-one-goal-p
;; 				       (cdr the-goals)
;; 				     (cdr goals))
;; 				   domain))

;; In case we want to translate a representation of the logistics or blocksworld in another one.
(defun change-names (problem domain)
 (sublis (case domain
	   ((logistics logistics-aips98-round2)
		      '((object . package)
			(post-office . location)
			(loc-at . in-city)
			(inside . in)))
	   (blocksworld '((object . block)
			  (on-table . ontable)
			  (arm-empty . handempty)))
	   (t nil))
	 problem))
