;; This allows to implement a best-first-search (A* or Dijkstra)
;; g values are computed from cost definitions at operators (or 1 if solution length is preferred)
;; h values are computed using a h-function. It is defined here, but it can be overriden in any domain definition

;; If heuristic computation requires knowing the state at each node, set *store-state-p* to t. 
;; It will be stored at (getf (p4::nexus-plist node) :state)
;; I also define :g-value and :h-value for each node

;;; IMPORTANT. This rule has to be defined in all files that are using control rules
;; Implements the A* procedure at each node decision point
;; (defun load-a-star-rule nil
;;   (control-rule CONTROL-NODE
;; 		(if (best-node <node>))
;; 		(then prefer node <node>)))

(in-package "USER")

(defun candidate-node (node)
  "Tests whether a node is among the candidate set of nodes in the
tree. Can be used as a generator."
  (declare (special *current-problem-space* *candidate-nodes*))
  (let ((candidates (p4::problem-space-property :expandable-nodes)))
    (cond ((p4::strong-is-var-p node)
	   (mapcar #'(lambda (cand) (list (cons node cand)))
		   candidates))
	  ((typep node 'p4::nexus)
	   (if (member node candidates) t))
	  ((typep node 'fixnum)
	   (if (member (find-node node) candidates) t)))))

;; It returns T if node is the best node according to the A* function. It can be used as generator
(defun best-node (node)
  "Tests whether a node is among the candidate set of nodes in the tree. It implements best-first search."
  (declare (special *current-problem-space*))
  (let ((best-candidate (best-candidate (p4::problem-space-property :expandable-nodes))))
    (cond ((p4::strong-is-var-p node)
	   (list (list (cons node best-candidate))))
	  ((typep node 'p4::nexus)
	   (if (eq node best-candidate) t))
	  ((typep node 'fixnum)
	   (if (eq (find-node node) best-candidate)
	       t)))))

;; It returns the best node according to the A* heuristic function from a set of nodes 
;; (trivial to convert it into returning a list of nodes)
;; First, it checks whether it is really close to the goal through the *lower-bound-h*
;; In case of tie in the F function, I prefer those with less H.
;; In case of tie in the H function, if both nodes are siblings, I prefer the oldest (first Prodigy alternative)
;; Otherwise, I prefer those newer in the search tree (kind of depth-first behavior)
(defun best-candidate (nodes)
  (when nodes
    (let* ((best-node (car nodes))
	   (g (compute-g best-node))
	   (h (compute-h best-node))
	   (h-best h)
	   (best-value (compute-f best-node g h)))
      (p4::output 3 t "~%Values of ~a are: ~a=~a+~a" best-node best-value g h)
      (dolist (node (cdr nodes) best-node)
	(let* ((g-node (compute-g node))
	       (h-node (compute-h node))
	       (value (compute-f node g-node h-node)))
	  (p4::output 3 t "~%Values of ~a are: ~a=~a+~a" node value g-node h-node)
	  (when (or (and (< h-node h-best)
			 (<= h-node *lower-bound-h*))
		    (< value best-value)
		    (and (= value best-value)
			 (or (< h-node h-best)
			     (and (= h-node h-best)
				  (if (eq (p4::nexus-parent node)
					  (p4::nexus-parent best-node))
				      (< (p4::nexus-name node) (p4::nexus-name best-node))
				    (> (p4::nexus-name node) (p4::nexus-name best-node)))))))
	    (setq best-node node)
	    (setq h-best h-node)
	    (setq best-value value)))))))

;; It is cached, given that this is a tree and not a graph
(defun compute-f (node g h)
  (cond ((getf (p4::nexus-plist node) :f-value))
	(t (setf (getf (p4::nexus-plist node) :f-value)
	     (+ g h)))))

;; If one does not want to cache the H value, because it is wanted to be
;; recomputed each time, then simply call h-function
(defun compute-h (node)
  (cond ((getf (p4::nexus-plist node) :h-value))
	(t (setf (getf (p4::nexus-plist node) :h-value)
	     (h-function node)))))

;; It computes and stores the g value for a node according to the selected
;; cost metric or number of steps (default)
;; It is cached, given that this is a tree and not a graph
(defun compute-g (node)
  (declare (special *cost-label* *current-node*))
  (cond ((getf (p4::nexus-plist node) :g-value))
	(t
	 (if (and *store-state-p*
		  (not (getf (p4::nexus-plist *current-node*) :state)))
	     (setf (getf (p4::nexus-plist node) :state)
	       (real-give-me-nice-state)))
	 (setf (getf (p4::nexus-plist node) :g-value)
	   (if (<= (p4::nexus-name node) 4)
	       0
	     (if (and (if *compute-f-at-bindings*
			  (p4::binding-node-p node)
			(p4::applied-op-node-p node))
		      (p4::costeable-operator-p node))
		 (+ (if (or (eq *cost-label* 'default)
			    (p4::additive-cost-p))
			(compute-g (p4::nexus-parent node))
		      0)
		    (p4::op-cost (p4::instantiated-op-op
				  (p4::a-or-b-node-instantiated-op node))
				 node))
	       (compute-g (p4::nexus-parent node))))))))


;; Simplest h function. It should be overriden
(defun h-function (node)
  (declare (ignore node))
  0)

;; An example of h function in the grid world
;; (defun compute-h (node)
;;   (let* ((initial-goals (give-me-nice-initial-goals))
;; 	 (current-state (or (getf (p4::nexus-plist node) :state)
;; 			    (setf (getf (p4::nexus-plist node) :state)
;; 			      (real-give-me-nice-state))))
;; 	 (current-goals (set-difference initial-goals current-state :test #'equal)))
;;     (apply #'+ (mapcar #'(lambda (goal)
;; 			   (distance goal current-state))
;; 		       current-goals))))

(in-package "P4")

(export '(define-if-not-prod-handler))

;; Defines a Prodigy handler in case it was not defined before

(defun define-if-not-prod-handler (condition function)
  (declare (symbol condition)
	   (function function)
	   (special *prodigy-handlers*))
  (let ((handlers (or (assoc condition *prodigy-handlers*)
		      (car (push (list condition) *prodigy-handlers*)))))
    (pushnew function (cdr handlers) :test #'equal))  
  )

;; Now, when previous control rule is used, it will compute the g value
;; (*this-cost*) If not, this will call the compute-g function to compute that
;; value
(defun best-first-search-handler (sig)
  (declare (ignore sig) 
	   (special user::*current-node* user::*best-cost* user::*this-cost*
		    user::*cost-label*))
  (let ((g-value (user::compute-g user::*current-node*)))
    (when (and (if user::*compute-f-at-bindings*
		   (binding-node-p user::*current-node*)
		 (applied-op-node-p user::*current-node*))
	       (costeable-operator-p user::*current-node*))
      (setf user::*this-cost* g-value)
      (when (and user::*backtrack-with-costs-p*
		 (<= user::*best-cost* user::*this-cost*))
	(output 2 t " Cost exceeds or equals best cost (~a).~%" user::*best-cost*)
	;;backtrack to some node
	(let ((new-node user::*current-node*))
	  (setf (getf (nexus-plist new-node) :termination-reason)
	    :exceeds-cost-bound)
	  (close-node-and-children new-node :exceeds-cost-bound))))))

;; Only if the best-first-search-handler wants to be used
(remove-prod-handler :always)
(define-prod-handler :always #'best-first-search-handler)


;; New one
(defun best-first-finish-handler (sig)
  (declare (ignore sig) 
	   (special user::*current-node* user::*best-cost* user::*this-cost*
		    user::*cost-label* user::*costs-declared*))
  (cond ((costeable-operator-p (car (nexus-children user::*current-node*)))
	 (let ((current-node (car (nexus-children user::*current-node*))))
	   (setf user::*this-cost* 
	     (+ (if (or (eq user::*cost-label* 'user::default)
			(additive-cost-p))
		    (user::compute-g (nexus-parent current-node))
		  0)
		(op-cost (instantiated-op-op
			  (applied-op-node-instantiated-op current-node))
			 current-node)))
	   (setf (getf (nexus-plist current-node) :g-value)
	     user::*this-cost*)
	   ;; Daniel changed this so that it does not provide equal cost solutions
	   (cond ((<= user::*this-cost* user::*best-cost*)
		  (setf user::*best-cost* user::*this-cost*)
		  (cons :stop :achieve))
		 (t (cond (user::*backtrack-with-costs-p*  
			   (output 2 t " Cost exceeds best cost (~a).~%" user::*best-cost*)
			   ;;backtrack to some node
			   (setf (getf (nexus-plist current-node) :termination-reason)
			     :exceeds-cost-bound)
			   (close-node-and-children current-node
						    :exceeds-cost-bound))
			  (t (cons :stop :achieve))))))) 
	;; Daniel changed this so that it does not provide equal cost solutions (it was <= before)
	(t (if (<= user::*this-cost* user::*best-cost*)
	       (setf user::*best-cost* user::*this-cost*))
	   (cons :stop :achieve))))

;; Only when one wants to use best-first-search
;; I set it as default
(remove-prod-handler :achieve)
(define-prod-handler :achieve #'best-first-finish-handler)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Nombre: is-op-in-rete
;;
;; Parametros:Ninguno
;; Valor devuelto: true,nil
;; Descripcion: Esta funcion detecta si un nodo ya esta incluido en la red 
;;    temporal.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun is-op-in-rete (op-in-rete)
  ;;Declaracion de variables.
  (declare (special user::*current-node*) 
	   )				;fin de la  declaracion

  (let* ((flag-rete nil)
	 (aux-ops (reverse user::*applied-operators-id*))
	 (list-null (null  user::*applied-operators-id*))
	 )				;fin de la declaracion de variables
     
    ;;Se recorre la lista applied-operators-id hasta se encuentre coincidencia o final de la lista
    ;;desde el ultimo elemento de la lista al primero.
    (loop while (and (not flag-rete) (not list-null) )
	do 
        
        
	  ;;Si coinciden los identificadores de los nodos....
	  (if (equal (p4::applied-op-node-name op-in-rete) (car aux-ops))
	      (progn
		(setq flag-rete t)
           
		)			;fin del progn
	    )				;fin del if
        
	  (setq aux-ops (cdr aux-ops))
         
	  ;;Si lista aux-ops es nil
	  (if (null aux-ops)
	      (setq list-null t)
	    )				;fin del if
         
        
	  );;fin del bucle     

    ;;Valor devuelto
    flag-rete
    )					;fin del let


  )					;fin de la funcion


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Nombre: get-num-applied-operators
;;
;; Parametros:list-diference
;; Valor devuelto: Numero de operadores aplicados.
;; Descripcion: Esta funcion devuelve el numero de operadores aplicados 
;;    de la lista que se pasa como parametro.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun get-num-applied-operators (list-diference)

  (let* ((num-op 0)
	 )				;fin de la declaracion de variables
     
    (loop for op in list-diference  
	do 
        
	  ;;Si es un operador aplicado,se incrementa el contador
	  (cond ((equal (type-of op)'PRODIGY4::APPLIED-OP-NODE)
		 (setq num-op (+ num-op 1)) 
		 )			;fin de si es op-applied
              
		)			;fin del cond      
	  );;fin del bucle que recorre los efectos     

    ;;Valor devuelto
    num-op
    )					;fin del let

  )					;fin de la funcion

;; (setf user::*use-quality-p* t)

(defun cost-ipss-handler-applied (sig)
  (declare (ignore sig) 
	   (special user::*current-node* user::*best-cost* user::*this-cost*
		    user::*cost-label*))
  

  (let* ((aux-path nil)
         (call-plan-handler t)		;flag usado para comprobar si ese operador hay que incluirlo o la rete o ya esta incluido
         (num 0)
         (aux-cost-label nil)
         (aux-duration nil)
         (aux-cost-function nil)
         
	 )				;fin de la declaracion de variables
   
    (when (and (applied-op-node-p user::*current-node*)
	       (costeable-operator-p user::*current-node*))
      (setf user::*this-cost*
	(+ (if (or (eq user::*cost-label* 'user::default)
		   (additive-cost-p))
	       (cost-at-backtracking-point
		(nexus-parent user::*current-node*))
	     0)
	   (op-cost (instantiated-op-op
		     (applied-op-node-instantiated-op user::*current-node*))
		    user::*current-node*)))
      (setf (getf (nexus-plist user::*current-node*) :g-value)
	user::*this-cost*) 
	  
	 
      ;;Si la metrica de coste no es time y user::*makespan* es t, se debe calcular el tiempo
      (cond ((and (not (equal user::*cost-label* 'time)) (problem-space-property :make-span))
                            
	     ;;Se calcula la duracion
	     ;;Se hace que el coste sea el tiempo (time)
	     (setq aux-cost-label user::*cost-label*)
	     (setf user::*cost-label* 'time)
	     (setq aux-cost-function
	       (choose-cost-function
		(instantiated-op-op (applied-op-node-instantiated-op user::*current-node*))
		user::*cost-label*)) 
	     ;;Si no existe funcion de coste de tiempo definida
	     (if (not aux-cost-function)       
		 (progn
		   (setf user::*cost-label* 'user::default)
					;(print 'no-tiene)(read t)
		   )			;fin del progn 
	       #|(progn 
                            ;;;; MD MARZO 2003 (print (costeable-operator-p (instantiated-op-op (applied-op-node-instantiated-op user::*current-node*))))
                            (setf user::*cost-label* 'time)
                            ;(print 'si-tiene)(read t)
                            );fin del prgn 
                            |#
	       )			;fin del if  
					;(print 'antes)   
                                                     
	     ;;Se calcula la duracion del operador.
	     (setf aux-duration
	       (+ (if (or (eq user::*cost-label* 'user::default)
			  (additive-cost-p))
		      (duration-at-backtracking-point
		       (nexus-parent user::*current-node*))
		    0)
		  (op-cost (instantiated-op-op
			    (applied-op-node-instantiated-op user::*current-node*))
			   user::*current-node*)))
	     (setf (getf (nexus-plist user::*current-node*) 'associated-duration)
	       aux-duration) 
					;(print aux-duration)(read t)
                                                       
	     ;;Se restaura el coste real
	     (setf user::*cost-label* aux-cost-label)
                            
	     )				;fin de la condicion1
    
	    )				;fin del cond
    
;;       (setf user::*backtrack-with-costs-p* t)
	 
      (when (and user::*backtrack-with-costs-p*
		 (< user::*best-cost* user::*this-cost*))
	(if user::*printp*
	    (output 2 t " Cost exceeds best cost (~a).~%" user::*best-cost*))
	;;backtrack to some node
	(let ((new-node user::*current-node*))
					;(break "before closing")
	  (setf (getf (nexus-plist new-node) :termination-reason)
	    :exceeds-cost-bound)
	  (close-node-and-children new-node :exceeds-cost-bound))))
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;;Se comprueba si se ha producido backtracking.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
					;(format stream "~%Cost-hanlder")
					;(print 'handler)
    (when (and (applied-op-node-p user::*current-node*) (problem-space-property :make-span))
      ;;Si es el primer operador operador aplicado, se almacena el path de este
      (cond (user::*first-op*
	     (setf user::*last-path* (p4::path-from-root user::*current-node*) )
                      
	     )				;fin de si es el primer operador
					;Else
	    (t 
	     ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
	     ;;producido backtracking
	     (cond ((<= (length (p4::path-from-root user::*current-node*))
			(length user::*last-path*) )
		    ;;Si este operador ya ha sido incluido en la rete, se modifica un flag para no llamar a plan-handler
		    (if (is-op-in-rete user::*current-node*)
			(progn
			  (setq call-plan-handler nil)
			  )		;fin del progn 
		      )			;fin del if
		    ;;Calculamos el num. de operadores aplicados a restaurar
		    (setq num (get-num-applied-operators
			       (set-difference user::*last-path*
					       (p4::path-from-root user::*current-node*))))
		    (restore-last-state num)
		    )			;fin de si es <=
					;Else
		   (t 
		    ;;Se comprueba si se ha producido backtraking..
		    ;;Obtenemos la lista del path actual de longitud = a la del path anterior
		    (setq aux-path 
		      (nthcdr (- (length (p4::path-from-root user::*current-node*))
				 (length user::*last-path*))
			      (reverse(p4::path-from-root user::*current-node*))))
                           
		    ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
		    (cond ((not (equal (reverse aux-path) user::*last-path*))
			   ;;Calculamos el num. de operadores aplicados a restaurar
			   (setq num (get-num-applied-operators
				      (set-difference user::*last-path*
						      (p4::path-from-root user::*current-node*))))
			   (restore-last-state num)
			   )		;fin de si son iguales
					;else = son iguales..
			  (t 
					;(print user::*applied-operators*) 
			   )		;fin del else
			  )		;fin del cond
		    )			;fin del else
		   )			;fin del cond 
	     ;;Se actualiza el path anterior con el actual
	     (setf user::*last-path* (p4::path-from-root user::*current-node*) )
	     )				;fin del else
	    )				;fin del cond
      (if call-plan-handler
	  ;;Se llama a plan-handler si el nodo actual es un nodo aplicado
	  (plan-handler) 	
	)				;fin del if
      )					;fin del when
   
    )					;fin del let
  
					;(print 'd)
 
  )					;fin del manejador


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Nombre: restore-last-state
;;
;; Parametros:Ninguno
;; Descripcion: Esta funcion elimina el operador aplicado anteriormente,
;;    el cual ha sido eliminado de la lista user::*applied-operators*, 
;;    el TP de dicho operador y su duracion de la lista user::*op-durations*.
;;    Tambien restaura el estado de la red temporal antes de haber aplicado
;,    ese operador.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun restore-last-state (aux-num)
  (if user::*last-op-consistence*
      ;;Se restaura el estado de la rte
      (progn
					;(print 'llamada1)
	(restore-rete-state aux-num)
	)
    ;;En caso contrario se restauran todos menos el ultimo xq no ha sido consistente.
    (progn
					;(print 'llamada2)
      (restore-rete-state (- aux-num 1))

      )
    )					;fin del if
					;(print 'llamada3)
  ;;Hay que restaurar estas variables tantas veces como operadores aplicados haya que quitar
  (loop while (> aux-num  0)
      do          
	(setf user::*applied-operators* (reverse(cdr(reverse user::*applied-operators*))) )
	(setf user::*applied-operators-id* (reverse(cdr(reverse user::*applied-operators-id*))) )
	(pop user::*list-TPs*) 
	(setf user::*op-durations* (reverse(cdr(reverse user::*op-durations*))) )
	(setf user::*abs-duration* (reverse(cdr(reverse user::*abs-duration*))) )
	(setq aux-num (- aux-num 1))
					;(print 'llamada4)		  
	)				;Fin del bucle que restaura el estado	  
					;(print 'llamada5)  
  )					;fin de la funcion 

(defun costeable-operator-p (node)
  (declare (special user::*costs-declared* user::*current-node*
		    user::*cost-label*))
  (let ((operator (instantiated-op-op (a-or-b-node-instantiated-op node))))
    (or (equal operator (get-operator *finish*))
	(eq user::*cost-label* 'user::default) 
	(and user::*costs-declared*
	     (choose-cost-function operator user::*cost-label*)))))

;;; Daniel. I do not understand why the operator is not used.
(defun additive-cost-p nil
  (declare (special user::*cost-label* *current-problem-space*))
  (member user::*cost-label*
	  (getf (problem-space-plist *current-problem-space*) 
		:additive-costs)))

;; Daniel. Not used.
;; (defun totally-binded-p (node operator)
;;   (declare (special user::*current-node*)
;; 	   (ignore node operator))
;;   (cdar (last (user::get-bindings-from-applied-node user::*current-node*))))


;; Only if IPSS
;; (remove-prod-handler :always)
;; (define-if-not-prod-handler :always #'cost-ipss-handler-applied)

(defun finish-ipss-handler (sig)
  (declare (ignore sig) 
	   (special user::*current-node* user::*best-cost* user::*this-cost*
		    user::*cost-label* user::*costs-declared*))
					;(print 'e)
  (let* ((aux-path nil)
         (op-name nil)
         (call-plan-handler t) 
         (num 0)
         (aux-duration 0)
         (aux-cost-label nil)
         (aux-cost-function nil)
         (actual (car (nexus-children user::*current-node*)))
	 )				;fin de la declaracion de variables
    (cond ((costeable-operator-p
	    (car (nexus-children user::*current-node*)))
					;(let ((actual (car (nexus-children user::*current-node*)))
					; );fin de la declaracion de variables
	   (setf user::*this-cost* 
	     (+ (if (or (eq user::*cost-label* 'user::default)
			(additive-cost-p))
		    (cost-at-backtracking-point (nexus-parent actual))
		  0)
		(op-cost (instantiated-op-op
			  (applied-op-node-instantiated-op actual))
			 actual)))
	   (setf (getf (nexus-plist actual) :g-value)
	     user::*this-cost*)
	   ;;Si la metrica de coste no es time y user::*make-span* es t, se debe calcular el tiempo
	   (cond ((and (not (equal user::*cost-label* 'time)) (problem-space-property :make-span))
		  ;;Se calcula la duracion
		  ;;Se hace que el coste sea el tiempo (time)
		  (setq aux-cost-label user::*cost-label*)
		  (setf user::*cost-label* 'time)
		  (setq aux-cost-function
		    (choose-cost-function
		     (instantiated-op-op (applied-op-node-instantiated-op actual))
		     user::*cost-label*)) 
		  ;;Si no existe funcion de coste de tiempo definida
		  (if (not aux-cost-function)       
		      (progn
			(setf user::*cost-label* 'user::default)
			)		;fin del progn 
		    )			;fin del if  
		  ;;Se calcula la duracion del operador.
		  (setf aux-duration
		    (+ (if (or (eq user::*cost-label* 'user::default)
			       (additive-cost-p))
			   (duration-at-backtracking-point
			    (nexus-parent actual))
			 0)
		       (op-cost (instantiated-op-op
				 (applied-op-node-instantiated-op actual))
				actual)))
		  (setf (getf (nexus-plist actual) 'associated-duration)
		    aux-duration) 
		  ;;Se restaura el coste real
		  (setf user::*cost-label* aux-cost-label)
		  )			;fin de la condicion1
		 )			;fin del cond
	   (cond ((<= user::*this-cost* user::*best-cost*)
		  ;; not here		  (setf user::*best-cost* user::*this-cost*)
		  ;;Si el make-span es t entonces se calcula el algoritmo integracion.
		  (cond ((problem-space-property :make-span)
			 ;;Si el padre es de tipo binding-node
			 (cond ((equal (p4::type-of (nexus-parent (car (nexus-children user::*current-node*))))
				       'PRODIGY4::BINDING-NODE)
				(setq op-name
				  (car (user::get-applicable-op
					(p4::binding-node-instantiated-op
					 (nexus-parent (car (nexus-children user::*current-node*)))) )) )
				;;Si es el operador finish, se pone el flag nil, para luego no mostrar la solucion
				(when (equal op-name 'PRODIGY4::*FINISH* )
				  (setf user::*no-finish* nil)
				  )	;fin del when 
				(when (not (equal op-name 'PRODIGY4::*FINISH* )) 
				  ;;Se comprueba si se ha producido backtracking.
				  (when (applied-op-node-p (car (nexus-children user::*current-node*)))
				    ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
				    ;;producido backtracking
				    (cond ((<= (length (p4::path-from-root (car (nexus-children user::*current-node*)))) 
					       (length user::*last-path*) ) 	
					   ;;Si este operador ya ha sido
					   ;;incluido en la rete, se modifica un flag para no llamar a plan-handler
					   (if (is-op-in-rete actual)
					       (progn
						 (setq call-plan-handler nil)
						 ) ;fin del progn 
					     ) ;fin del if
					   ;;Calculamos el num. de operadores aplicados a restaurar
					   (setq num 
					     (get-num-applied-operators
					      (set-difference user::*last-path*
							      (p4::path-from-root (car (nexus-children user::*current-node*))))))
					   (restore-last-state num)
					   ) ;fin de si es <=
					;Else
					  (t 
					   ;;Se comprueba si se ha producido backtraking..
					   ;;Obtenemos la lista del path actual de longitud = a la del path anterior
					   (setq aux-path
					     (nthcdr (- (length (p4::path-from-root (car (nexus-children user::*current-node*))))
							(length user::*last-path*))
						     (reverse (p4::path-from-root (car (nexus-children user::*current-node*)))) ))
					   ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
					   (cond ((not (equal (reverse aux-path) user::*last-path*)) 
						  ;;Calculamos el num. de operadores aplicados a restaurar
						  (setq num 
						    (get-num-applied-operators 
						     (set-difference user::*last-path*
								     (p4::path-from-root
								      (car (nexus-children user::*current-node*))))))
						  (restore-last-state num)
						  ) ;fin de si son iguales
					;else = son iguales..
						 (t ;(print 'iguales)
						  ) ;fin del else
						 ) ;fin del cond
					   ) ;fin del else
					  ) ;fin del cond 
				    ;;Se actualiza el path anterior con el actual
				    (setf user::*last-path* (p4::path-from-root (car (nexus-children user::*current-node*))) )
				    (if call-plan-handler
					(progn
					  ;;Se llama a finish-plan-handler si el nodo actual es un nodo aplicado
					  (finish-plan-handler) 
					  ) ;fin del progn
				      )	;fin del if
				    )	;fin del when  
				  )	;fin del when
				)	;fin del cond
					;else
			       (t ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
				;;Se comprueba si se ha producido backtracking.
                                ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
				(when (and (applied-op-node-p (car (nexus-children user::*current-node*)))
					   user::*last-op-consistence* )
				  ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
				  ;;producido backtracking
				  (cond ((<= (length (p4::path-from-root (car (nexus-children user::*current-node*))))
					     (length user::*last-path*) ) 
					 ;;Si este operador ya ha sido
					 ;;incluido en la rete, se modifica un flag para no llamar a plan-handler
					 (if (is-op-in-rete actual)
					     (progn
					       (setq call-plan-handler nil)
					       ) ;fin del progn 
					   ) ;fin del if
					 ;;Calculamos el num. de operadores aplicados a restaurar
					 (setq num
					   (get-num-applied-operators
					    (set-difference user::*last-path*
							    (p4::path-from-root (car (nexus-children user::*current-node*))))))
					 (restore-last-state num)
					 ) ;fin de si es <=
					;Else
					(t 
					 ;;Se comprueba si se ha producido backtraking..
					 ;;Obtenemos la lista del path actual de longitud = a la del path anterior
					 (setq aux-path
					   (nthcdr (- (length (p4::path-from-root (car (nexus-children user::*current-node*))))
						      (length user::*last-path*))
						   (reverse (p4::path-from-root
							     (car (nexus-children user::*current-node*)))) ))
					 ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
					 (cond ((not (equal (reverse aux-path) user::*last-path*)) 
						;;Calculamos el num. de operadores aplicados a restaurar
						(setq num
						  (get-num-applied-operators
						   (set-difference user::*last-path*
								   (p4::path-from-root
								    (car (nexus-children user::*current-node*))))))
						(restore-last-state num)
						) ;fin de si son iguales
					;else = son iguales..
					       (t ;(print 'iguales)
                                    
						) ;fin del else
					       ) ;fin del cond
					 ) ;fin del else
					) ;fin del cond 
				  ;;Se actualiza el path anterior con el actual
				  (setf user::*last-path* (p4::path-from-root (car (nexus-children user::*current-node*))) )
				  (if call-plan-handler
				      (progn
					;;Se llama a finish-plan-handler si el nodo actual es un nodo aplicado
					(finish-plan-handler) 
					) ;fin del progn
				    )	;fin del if
				  )	;fin del when  
				)	;fin del else(distinto de *finish*) 
			       )	;find el cond
			 )		;fin de la condicion de si :make-span = t
			)		;find el cond
		  ;;(break "primer lugar")
		  ;;Si la rete es consistente... entonces finaliza la busqueda de una solucion.
		  (cond ((and user::*consistence* user::*last-op-consistence*)    
			 (setf user::*best-cost* user::*this-cost*)			
			 (cons :stop :achieve)
		         )		;fin de si es consistente
					;Else
			(t 
			 (setf user::*consistence* t)
		         )
			)		;fin del cond
		  )
		 ;; *this-cost* no es menor que el mejor coste 
		 (t (cond (user::*backtrack-with-costs-p*  
			   (if user::*printp*
			       (output 2 t " Cost exceeds best cost (~a).~%" 
				       user::*best-cost*))
			   ;;backtrack to some node
			   (setf (getf (nexus-plist actual)
				       :termination-reason)
			     :exceeds-cost-bound)
			   (close-node-and-children actual
						    :exceeds-cost-bound))
			  (t 
			   (cond ((problem-space-property :make-span)  
				  ;;Si el padre es de tipo binding-node
				  (cond ((equal (p4::type-of (nexus-parent (car (nexus-children user::*current-node*))))
						'PRODIGY4::BINDING-NODE)
					 (setq op-name
					   (car (user::get-applicable-op 
						 (p4::binding-node-instantiated-op 
						  (nexus-parent
						   (car (nexus-children user::*current-node*)))) )) )
					 ;;Si es el operador finish, se pone el flag nil, para luego no mostrar la solucion
					 (when (equal op-name 'PRODIGY4::*FINISH* )
					   (setf user::*no-finish* nil)
					   ) ;fin del when 
					 (when (not (equal op-name 'PRODIGY4::*FINISH* )) 
					   ;;Se comprueba si se ha producido backtracking.
					   (when (applied-op-node-p (car (nexus-children user::*current-node*)))
					     ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
					     ;;producido backtracking
					     (cond ((<= (length (p4::path-from-root (car (nexus-children user::*current-node*))))
							(length user::*last-path*) ) 	
						    ;;Si este operador ya ha
						    ;;sido incluido en la rete, se modifica un flag para no llamar a plan-handler
						    (if (is-op-in-rete actual)
							(progn
							  (setq call-plan-handler nil)
							  ) ;fin del progn 
						      ) ;fin del if
						    ;;Calculamos el num. de operadores aplicados a restaurar
						    (setq num
						      (get-num-applied-operators
						       (set-difference user::*last-path*
								       (p4::path-from-root
									(car (nexus-children user::*current-node*))))))
						    (restore-last-state num)
						    ) ;fin de si es <=
					;Else
						   (t 
						    ;;Se comprueba si se ha producido backtraking..
						    ;;Obtenemos la lista del path actual de longitud = a la del path anterior
						    (setq aux-path
						      (nthcdr (- (length (p4::path-from-root
									  (car (nexus-children user::*current-node*))))
								 (length user::*last-path*))
							      (reverse (p4::path-from-root
									(car (nexus-children user::*current-node*)))) ))
						    ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
						    (cond ((not (equal (reverse aux-path) user::*last-path*)) 
							   ;;Calculamos el num. de operadores aplicados a restaurar
							   (setq num 
							     (get-num-applied-operators
							      (set-difference user::*last-path*
									      (p4::path-from-root
									       (car (nexus-children user::*current-node*))))))
							   (restore-last-state num)
							   ) ;fin de si son iguales
					;else = son iguales..
							  (t ;(print 'iguales)
							   ) ;fin del else
							  ) ;fin del cond
						    ) ;fin del else
						   ) ;fin del cond 
					     ;;Se actualiza el path anterior con el actual
					     (setf user::*last-path*
					       (p4::path-from-root (car (nexus-children user::*current-node*))) )
					     (if call-plan-handler
						 (progn
						   ;;Se llama a finish-plan-handler si el nodo actual es un nodo aplicado
						   (finish-plan-handler) 
						   ) ;Fin del progn
					       ) ;fin del if
					     ) ;fin del when  
					   ) ;fin del when
					 ) ;fin del cond
					;else
					(t
					 ;;Se comprueba si se ha producido backtracking.
					 (when (and (applied-op-node-p (car (nexus-children user::*current-node*)))
						    user::*last-op-consistence* )
					   ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
					   ;;producido backtracking
					   (cond ((<= (length (p4::path-from-root (car (nexus-children user::*current-node*))))
						      (length user::*last-path*) ) 	
						  ;;Si este operador ya ha
						  ;;sido incluido en la rete, se modifica un flag para no llamar a plan-handler
						  (if (is-op-in-rete actual)
						      (progn
							(setq call-plan-handler nil)
							) ;fin del progn 
						    ) ;fin del if
						  ;;Calculamos el num. de operadores aplicados a restaurar
						  (setq num
						    (get-num-applied-operators
						     (set-difference user::*last-path*
								     (p4::path-from-root
								      (car (nexus-children user::*current-node*))))))
						  (restore-last-state num)
						  ) ;fin de si es <=
					;Else
						 (t 
						  ;;Se comprueba si se ha producido backtraking..
						  ;;Obtenemos la lista del path actual de longitud = a la del path anterior
						  (setq aux-path
						    (nthcdr (- (length (p4::path-from-root
									(car (nexus-children user::*current-node*))))
							       (length user::*last-path*))
							    (reverse (p4::path-from-root
								      (car (nexus-children user::*current-node*)))) ))
						  ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
						  (cond ((not (equal (reverse aux-path) user::*last-path*)) 
							 ;;Calculamos el num. de operadores aplicados a restaurar
							 (setq num
							   (get-num-applied-operators
							    (set-difference user::*last-path*
									    (p4::path-from-root
									     (car (nexus-children user::*current-node*))))))
							 (restore-last-state num)
							 ) ;fin de si son iguales
					;else = son iguales..
							(t ;(print 'iguales)
							 ) ;fin del else
							) ;fin del cond
						  ) ;fin del else
						 ) ;fin del cond 
					   ;;Se actualiza el path anterior con el actual
					   (setf user::*last-path*
					     (p4::path-from-root (car (nexus-children user::*current-node*))) )
					   (if call-plan-handler
					       (progn
						 ;;Se llama a finish-plan-handler si el nodo actual es un nodo aplicado
						 (finish-plan-handler) 
						 ) ;fin del progn
					     ) ;fin del if
					   ) ;fin del when  
					 ) ;fin del else(distinto de *finish*) 
					) ;find el cond
				  )	;fin de la cond de si :make-span es t
				 )	;fin del cond    
			   ;;Si la rete es consistente... entonces finaliza la busqueda de una solucion.
			   (cond ((and user::*consistence* user::*last-op-consistence*)   
				  (setf user::*best-cost* user::*this-cost*)
				  (cons :stop :achieve)
				  )	;fin de si es consistente
					;Else
				 (t 
				  (setf user::*consistence* t)
				  )
				 )	;fin del cond
			   )))))	;) 
	  (t 
	   ;;Si el make-span = t se calcula el algoritmo
	   (cond ((problem-space-property :make-span)
	      
		  ;;Si el padre es de tipo binding-node
		  (cond ((eq (p4::type-of (nexus-parent (car (nexus-children user::*current-node*))))
			     'PRODIGY4::BINDING-NODE)
			 (setq op-name
			   (car (user::get-applicable-op
				 (p4::binding-node-instantiated-op
				  (nexus-parent (car (nexus-children user::*current-node*)))) )) )
			 ;;Si es el operador finish, se pone el flag nil, para luego no mostrar la solucion
			 (when (equal op-name 'PRODIGY4::*FINISH* )
			   (setf user::*no-finish* nil)
			   )		;fin del when 
			 (unless (eq op-name 'PRODIGY4::*FINISH* )
			   ;;Se comprueba si se ha producido backtracking.
			   (when (applied-op-node-p (car (nexus-children user::*current-node*)))
			     ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
			     ;;producido backtracking
			     (cond ((<= (length (p4::path-from-root
						 (car (nexus-children user::*current-node*))))
					(length user::*last-path*) )    	
				    ;;Si este operador ya ha sido incluido en
				    ;;la rete, se modifica un flag para no llamar a plan-handler
				    (if (is-op-in-rete actual)
					(progn
					  (setq call-plan-handler nil)
					  ) ;fin del progn 
				      )	;fin del if
				    ;;Calculamos el num. de operadores aplicados a restaurar
				    (setq num
				      (get-num-applied-operators
				       (set-difference user::*last-path*
						       (p4::path-from-root (car (nexus-children user::*current-node*))))))
				    (restore-last-state num)
				    )	;fin de si es <=
					;Else
				   (t 
				    ;;Se comprueba si se ha producido backtraking..
				    ;;Obtenemos la lista del path actual de longitud = a la del path anterior
				    (setq aux-path
				      (nthcdr (- (length (p4::path-from-root (car (nexus-children user::*current-node*))))
						 (length user::*last-path*))
					      (reverse(p4::path-from-root (car (nexus-children user::*current-node*)))) ))
				    ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
				    (cond ((not (equal (reverse aux-path) user::*last-path*))   
					   ;;Calculamos el num. de operadores aplicados a restaurar
					   (setq num 
					     (get-num-applied-operators
					      (set-difference user::*last-path*
							      (p4::path-from-root (car (nexus-children user::*current-node*))))))
					   (restore-last-state num)
					   ) ;fin de si son iguales
					;else = son iguales..
					  (t ;(print 'iguales)
					   ) ;fin del else
					  ) ;fin del cond
				    )	;fin del else
				   )	;fin del cond 
			     ;;Se actualiza el path anterior con el actual
			     (setf user::*last-path* (p4::path-from-root (car (nexus-children user::*current-node*))) )
			     (if  call-plan-handler
				 ;;Se llama a finish-plan-handler si el nodo actual es un nodo aplicado
				 (finish-plan-handler) 
			       )	;fin del if
			     )		;fin del when  
			   )		;fin del when
			 )		;fin del cond
					;else
			(t
			 ;;Se comprueba si se ha producido backtracking.
			 (when (and (applied-op-node-p (car (nexus-children user::*current-node*)))
				    user::*last-op-consistence* )
			   ;;Si la longitud del path actual es <= que la del anterior, es que se ha 
			   ;;producido backtracking
			   (cond ((<= (length (p4::path-from-root (car (nexus-children user::*current-node*))))
				      (length user::*last-path*) ) 
				  ;;Si este operador ya ha sido incluido en la
				  ;;rete, se modifica un flag para no llamar a plan-handler
				  (if (is-op-in-rete actual)
				      (progn
					(setq call-plan-handler nil)
					) ;fin del progn 
				    )	;fin del if
				  ;;Calculamos el num. de operadores aplicados a restaurar
				  (setq num (get-num-applied-operators
					     (set-difference user::*last-path*
							     (p4::path-from-root (car (nexus-children user::*current-node*))))))
				  (restore-last-state num)
				  )	;fin de si es <=
					;Else
				 (t 
				  ;;Se comprueba si se ha producido backtraking..
				  ;;Obtenemos la lista del path actual de longitud = a la del path anterior
				  (setq aux-path 
				    (nthcdr (- (length (p4::path-from-root (car (nexus-children user::*current-node*))))
					       (length user::*last-path*))
					    (reverse (p4::path-from-root (car (nexus-children user::*current-node*)))) ))
				  ;;Si no coinciden aux-path y el path anterior... entonces ha habido backtracking.
				  (cond ((not (equal (reverse aux-path) user::*last-path*)) 
					 ;;Calculamos el num. de operadores aplicados a restaurar
					 (setq num
					   (get-num-applied-operators
					    (set-difference user::*last-path*
							    (p4::path-from-root (car (nexus-children user::*current-node*))))))
					 (restore-last-state num)
					 ) ;fin de si son iguales
					;else = son iguales..
					(t ;(print 'iguales)
					 ) ;fin del else
					) ;fin del cond
				  )	;fin del else
				 )	;fin del cond 
			   ;;Se actualiza el path anterior con el actual
			   (setf user::*last-path* (p4::path-from-root (car (nexus-children user::*current-node*))) )
			   (if call-plan-handler
			       (progn
				 ;;Se llama a finish-plan-handler si el nodo actual es un nodo aplicado
				 (finish-plan-handler) 
				 )	;fin
			     )		;fin del if
			   )		;fin del when  
			 )		;fin del else 
			)		;find el cond
		  )			;fin de la cond del make-span = t
		 )			;fin del cond 
	   ;;(break "tercer lugar")
	   ;;Si la rete es consistente... entonces finaliza la busqueda de una solucion.
	   (cond ((and user::*consistence* user::*last-op-consistence*)    
		  (when (<= user::*this-cost* user::*best-cost*)
		    (setf user::*best-cost* user::*this-cost*))
		  (cons :stop :achieve)
		  )			;fin de si es consistente
					;Else
		 (t  
		  (setf user::*consistence* t)
		  )
		 )			;fin del cond
	   ))
    )					;fin del let
					;(print 'e)
  )					;fin del manejador

;; Only if IPSS
;; (remove-prod-handler :achieve)
;; (define-if-not-prod-handler :achieve #'finish-ipss-handler)

(defun cost-at-backtracking-point (node)
  (cond ((null (nexus-parent node)) 0)
	((and (applied-op-node-p node)
	      (getf (nexus-plist node) :g-value)))
	(t (cost-at-backtracking-point (nexus-parent node)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Nombre: duration-at-backtracking-point
;; Parametros: node
;; Valor devuelto: La duracion hasta el nodo actual
;; Descripcion:
;;     Esta funcion calcula la duracion hasta el nodo actual.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun duration-at-backtracking-point (node)
  (cond ((null (nexus-parent node)) 0)
	((and (applied-op-node-p node)
	      (getf (nexus-plist node) 'associated-duration)))
	(t (duration-at-backtracking-point (nexus-parent node)))))


(defun op-cost (operator node)
  (declare (special user::*cost-label*))
  (cond ((equal operator (get-operator *finish*))
	 0)
	((eq user::*cost-label* 'user::default)
	 1)
	(t (let ((cost-function (choose-cost-function operator user::*cost-label*)))
	     (if (numberp cost-function)
		 cost-function
	       (eval (sublis (get-bindings-from-applied-node node) cost-function)))))))

;; Returns the appropriate cost metric definition from the ones stored in the operator
(defun choose-cost-function (operator cost-metric)
  (some #'(lambda (op-cost-metric)
	    (if (eq (car op-cost-metric) cost-metric)
		(second op-cost-metric)))
	(second (getf (operator-plist operator) :costs))))

(defun make-my-list (vars values)
  (if (and values
	   (car values))
      (cons (cons (car vars) (car values))
	    (make-my-list (cdr vars) (cdr values)))))

(defun ligadas (list)
  (let ((res nil))
    (dolist (aux list res)
      (if (cdr aux)
	  (push (car aux) res)))))