;;; This code takes a problem file in PDL format and generates as output a
;;; path planning problem in a suitable format for our search library.

;; translator-rules have the format:
;; ((?pdl-node-predicate-x pdl-node-predicate-x ?node-name-argument ?node-x-argument)
;;  (?pdl-node-predicate-y pdl-node-predicate-y ?node-name-argument ?node-y-argument)
;;  (?pdl-obstacle-predicate-x pdl-obstacle-predicate-x ?obstacle-name-argument ?obstacle-x-argument)
;;  (?pdl-obstacle-predicate-y pdl-obstacle-predicate-y ?obstacle-name-argument ?obstacle-y-argument)
;;  (?pdl-arc-predicate pdl-arc-predicate ?origin-argument ?destination-argument)
;;  (?pdl-arc-predicate-cost pdl-arc-predicate-cost 'node-argument))

(defun translate-tp-to-pp (problem pp-file 
			   &key (translator-rules *pp-translator-rules*)
				(fill-arcs-fn #'adjacent-node-p))
  (declare (special *pp-translator-rules*))
  (let (rule coordinates node arcs obstacle obstacles state)
    (cond ((and (stringp problem) (probe-file problem))
	   (with-open-file (istream problem :direction :input)  
	     (setq state (cdadr (cadddr (caddr (read istream)))))))
	  (t (setq state (cdadr (cadddr (caddr problem))))))
    (dolist (literal state)
      (when (setq rule (car (member (car literal) translator-rules :key #'car :test #'eq)))
	;; 	    (format t "~%~a ~a" rule literal)
	(case (cadr rule)
	  (pdl-node-predicate-x 
	   (if (setq node (cdr (assoc (nth (caddr rule) literal)
				      coordinates)))
	       (push (nth (cadddr rule) literal)
		     (cdr (assoc (nth (caddr rule) literal)
				 coordinates)))
	     (push (list (nth (caddr rule) literal)
			 (nth (cadddr rule) literal))
		   coordinates)))
	  (pdl-obstacle-predicate-x 
	   (if (cdr (assoc (nth (caddr rule) literal)
			   obstacles))
	       (push (nth (cadddr rule) literal)
		     (cdr (assoc (nth (caddr rule) literal)
				 obstacles)))
	     (push (list (nth (caddr rule) literal)
			 (nth (cadddr rule) literal))
		   obstacles)))
	  (pdl-node-predicate-y 
	   (if (setq node (assoc (nth (caddr rule) literal)
				 coordinates))
	       (setf (cdr node) 
		 (list (cadr node)
		       (nth (cadddr rule) literal)))
	     (push (list (nth (caddr rule) literal)
			 (nth (cadddr rule) literal))
		   coordinates)))
	  (pdl-obstacle-predicate-y 
	   (if (setq obstacle (assoc (nth (caddr rule) literal)
				     obstacles))
	       (setf (cdr obstacle) 
		 (list (cadr obstacle)
		       (nth (cadddr rule) literal)))
	     (push (list (nth (caddr rule) literal)
			 (nth (cadddr rule) literal))
		   obstacles)))
	  (pdl-arc-predicate
	   (push (list (nth (caddr rule) literal)
		       (nth (cadddr rule) literal))
		 arcs))
	  (pdl-arc-predicate-cost)
	  )))
    (if fill-arcs-fn
	(dolist (node1 coordinates)
	  (dolist (node2 coordinates)
	    (if (and (not (member (cdr node2) obstacles :key #'cdr :test #'equal))
		     (funcall fill-arcs-fn node1 node2))
		(push (list (car node1) (car node2))
		      arcs)))))
    (with-open-file (ostream pp-file :direction :output
		     :if-exists :supersede :if-does-not-exist :create)
      (format ostream "(setf *nodes*~%   '(")
      (dolist (node coordinates)
	(format ostream "~%       ~((coordinates ~a ~a ~a)~)"
		(car node) (cadr node) (caddr node)))
      (format ostream "~%))~2%(setf *arcs*~%   '(")
      (dolist (node arcs)
	(format ostream "~%       ~((arc ~a ~a walk walk 1 1 1 1)~)"
		(car node) (cadr node) (caddr node)))
      (format ostream
	      "~%))~2%(setf *connection-array* (create-connection-array *nodes* *arcs*))"))))

;; It returns T if both nodes are adjacent
(defun adjacent-node-p (node1 node2)
  (and (not (equal node1 node2))
       (<= (distance (cadr node1) (cadr node2) (caddr node1) (caddr node2))
	   1)))

