;;; **************************************************
;;;           Utility-related functions
;;; **************************************************
(in-package "USER")

;; Allows to compute an estimation of the utility of control rules so that the
;; most expensive ones can be removed

;; It removes rules incrementally by analizing its utility
(defun incremental-utility-pruning (rule-data ostream)
  (dolist (rule *cr-list*)
    (let* ((total-matching-time (nth 1 rule-data))
	   (matchings (nth 2 rule-data))
	   (matching-time (if (> matchings 0)
			      (/ total-matching-time
				 matchings)
			      most-positive-fixnum))
	   (num-nodes (nth 3 rule-data))
	   (fires (nth 4 rule-data))
	   (le-steps (nth 5 rule-data))
	   (le-oldness (nth 6 rule-data)))
;; (rule-name (p4::control-rule-name rule))
;; 	   (le (gethash rule-name *cr-to-le-hash*))
      (if (and rule
	       (numberp matchings)
	       (numberp matching-time)
	       (numberp num-nodes)
	       (numberp fires)
	       (numberp le-steps)
	       (numberp le-oldness)
	       (usefulp matchings matching-time num-nodes fires
			le-steps le-oldness))
	  (save-prodigy-cr rule ostream)))))

;; It stores the rules in rules-file into output-file such that they have
;; positive utility
;; Asking for the existence of rule is required because after saving
;; utility data, rules might be refined, and some rules might disappear

(defun prune-rules-utility (domain &key (data-file *utility-output*)
				   (rules-file *cr-file*)
				   (output-file "pruned-rules"))
  (reset-domain-variables nil domain)
  (with-open-file (ostream (format nil "~a~a-~d.~a" *domain-directory*
				   output-file *utility-threshold*
				   *lisp-extension*)
			   :direction :output
			   :if-exists :supersede
			   :if-does-not-exist :create)
    (cond ((probe-file data-file)
	   (domain domain :load-domain nil :function :compile)
	   (setf *cr-file* rules-file)
	   (load (cr-file))
	   (load-domain)
	   (load data-file)
;; 	   (let ((rules (all-control-rules)))
	   (dolist (rule-data *utility-data*)
	     (let* ((rule-name (nth 0 rule-data))
		    (total-matching-time (nth 1 rule-data))
		    (matchings (nth 2 rule-data))
		    (matching-time (if (> matchings 0)
				       (/ total-matching-time
					  matchings)
				       most-positive-fixnum))
		    (num-nodes (nth 3 rule-data))
		    (fires (nth 4 rule-data))
		    (le-steps (nth 5 rule-data))
		    (le-oldness (nth 6 rule-data))
		    (rule (from-name-to-rule rule-name)))
	       (if (and rule
			(numberp matchings)
			(numberp matching-time)
			(numberp num-nodes)
			(numberp fires)
			(numberp le-steps)
			(numberp le-oldness)
			(usefulp matchings matching-time num-nodes fires
				 le-steps le-oldness))
		   (save-prodigy-cr rule ostream)))))
	  (t nil))))

(defun usefulp (matchings matching-time num-nodes fires le-steps le-oldness)
  (declare (ignore le-steps le-oldness))
  (and (> fires 1)
       (> (utility-count matchings matching-time num-nodes fires)
	  *utility-threshold*)))

(defun utility-count (matchings matching-time num-nodes fires)
  (- (* num-nodes
	(/ fires (if (> matchings 0)
		     matchings
		     most-positive-fixnum))
	*time-per-node*)
     matching-time))

;;
;; Inits the structures for the utility analysis only when there are no les.
;;
(DEFUN init-for-utility (&optional (rules (all-control-rules)))
  (if (and rules
	   (not (gethash (p4::control-rule-name (car rules)) *cr-to-le-hash*)))
      (dolist (rule rules)
	(setf (getf (p4::control-rule-plist rule) :matching-time) 0)
	(setf (getf (p4::control-rule-plist rule) :matchings) 0)
	(setf (getf (p4::control-rule-plist rule) :num-nodes) 0)
	(setf (getf (p4::control-rule-plist rule) :firings) 0))))

#|
(DEFUN init-for-utility (&optional (rules (all-control-rules)))
  (if rules
      (cond ((gethash (p4::control-rule-name (car rules)) *cr-to-le-hash*)
	     (maphash #'(lambda (key le)
			  (declare (ignore key))
			  (setf (learning-episode-matching-time le) 0)
			  (setf (learning-episode-matchings le) 0)
			  (setf (learning-episode-firings le) 0))
		      *cr-to-le-hash*))
	    (t (dolist (rule rules)
		 (setf (getf (p4::control-rule-plist rule) :matching-time) 0)
		 (setf (getf (p4::control-rule-plist rule) :matchings) 0)
		 (setf (getf (p4::control-rule-plist rule) :firings) 0))))))
|#


;;
;; Some pprint function
;;
(DEFUN pp-utility (&optional (rules (all-control-rules)) (stream t))
  (let ((in-display-p (eq stream t)))
    (if in-display-p
	(format stream "~2%~20tRULE~60tTIME~67tMATCH~73tNUM-NODES~78tFIRES STEPS LE-TIME"))
    (dolist (rule rules)
      (let* ((rule-name (p4::control-rule-name rule))
	     (le (gethash rule-name *cr-to-le-hash*)))
	(cond (le
	       (format stream (if in-display-p
				  "~%~(~a~)~60t~,5f~66t~5d~72t~5d~77t~5d~85t~5d~90t~,5f"
				  "~%~7t(~(~a~) ~f ~d ~d ~d ~d ~,5f)")
		       rule-name
		       (learning-episode-matching-time le)
		       (learning-episode-matchings le)
		       (learning-episode-num-nodes le)
		       (learning-episode-firings le)
		       (learning-episode-learning-steps le)
		       (* 100.0
			  (/ (read-from-string
			      (string-left-trim '(#\L #\E #\l #\e #\-)
						(string
						 (learning-episode-name le))))
			     (1+ *le-index*)))))
	      (t (format stream (if in-display-p
				  "~%~(~a~)~60t~,5f~66t~5d~72t~5d~77t~5d0 0"
				  "~%~7t(~(~a~) ~f ~d ~d ~d 0 0)")
			 rule-name
			 (getf (p4::control-rule-plist rule) :matching-time)
			 (getf (p4::control-rule-plist rule) :matchings)
			 (getf (p4::control-rule-plist rule) :num-nodes)
			 (getf (p4::control-rule-plist rule) :firings))))))))

;;
;; Saving the utility information, plus the number of steps to learn, and the
;; time it learn the rules, so that we can perform an utility analysis
;;
(DEFUN save-utility (&optional (u-file *utility-output*))
  (with-open-file (file u-file :direction :output
			:if-exists :supersede
			:if-does-not-exist :create)
    (format file
	    ";;;RULE-NAME TIME-TO-MATCH MATCHINGS NUM-NODES FIRES LE-STEPS LE-TIME")
    (format file "~2%(setf *utility-data* '(")
    (pp-utility (all-control-rules) file)
    (format file "~%))~%")))
