;; SAYPHI Case-Based Reasoner
;; General Functions for Case Based Planning
;; Td. 10.02.2006
;; __________________________________________________________________________________

(defvar *replay-table* nil)
(defvar *type-case-base* nil)
(defvar *object-case-base* nil)

;; Macros sustitutivas para la etapa de Replay
(defmacro object-replay (object)
  `(gethash ,object *replay-table*))

(defmacro get-case-step (object)
  `(let ((obreplay (object-replay ,object)))
     (nth (+ 1 (replay-case-pointer obreplay)) (replay-sequence obreplay))))

(defmacro get-case-step-ahead (object)
  `(let ((obreplay (object-replay ,object)))
     (nth (+ 2 (replay-case-pointer obreplay)) (replay-sequence obreplay))))


(defmacro get-nth-case-step (object pointer &optional (ahead 1))
  `(nth (+ ,ahead ,pointer) (replay-sequence (object-replay ,object))))


(defmacro markused-case (object)
  `(setf (replay-used-in-step (object-replay ,object)) t))
	
(defmacro advance-type-case (object)
  `(progn (incf (replay-case-pointer (object-replay ,object)))
	  (setf (replay-no-op-counter (object-replay ,object)) 0))) 

(defmacro node-objcase-pointer (node object)
  `(gethash ,object (snode-replay-pointers ,node)))

;; ===============================================================================


(defmacro equal-substate (step-case1 step-case2)
 `(and (equal (casestep-properties ,step-case1) (casestep-properties ,step-case2))
       (equal (casestep-prop-extras ,step-case1) (casestep-prop-extras ,step-case2))))

(defmacro equal-app-action (step-case1 step-case2)
  `(eq (car (casestep-app-action ,step-case1)) (car (casestep-app-action ,step-case2))))

(defmacro noop-step-p (step)
  `(and (casestep-p ,step)
	(eq (car (casestep-app-action ,step)) 'no-op)))

(defmacro equal-stepaction (stepaction xaction)
   `(eq (car ,stepaction) (car ,xaction)))



;; To compute states I use a different state represetation called pstate from 
;; propositional state.  In this case the hash table contains pred as key and
;; a list of objects literals as values
(defun get-pstate(state)
  (let ((predicates (dom-predicates *pspace*))
	(objects (problem-inheritobjects *current-problem*))
	(pstate (make-hash-table :test #'eq)))
    (maphash #'(lambda (pred bitmap)
		 (setf (gethash pred pstate)
		       (realinstances-from-bitmap bitmap (gethash pred predicates) objects)))
	     state)
    pstate))

;;Given a list of real objects, this function return in a list all 
;; positions in which the object is present
(defun object-positions (object obj-inst)
  (let ((positions nil))
    (dotimes (i (length obj-inst) (reverse positions))
      (when (eq object (nth i obj-inst))
	(push i positions)))))
    

(defun get-all-types (domtypes)
  (let ((all-types nil))
    (dolist (itype domtypes all-types)
      (cond ((listp itype)
	     (unless (find (car itype) all-types)
	       (push (car itype) all-types))
	     (dolist (i-child (cadr itype))
	       (unless (find i-child all-types)
		 (push i-child all-types))))
	    (t
	     (unless (find itype all-types)
	       (push itype all-types)))))))


(defun is-object-type (ob-inst type)
  (let ((objects (problem-inheritobjects *current-problem*)))
    (when (find ob-inst (gethash type objects)) t)))



(defun get-all-objects ()
  (let  ((all-objects nil))
    (maphash #'(lambda (type objects)
		 (declare (ignore type))
		 (dolist (i-ob objects)
		   (push i-ob all-objects)))
	     (problem-objects *current-problem*))
    all-objects
    ))


(defun stepliterals-memberset (case-literals xliterals)
  (let ((is-memberset t))
    (dolist (i-lit xliterals is-memberset)
      (when is-memberset
	(unless (find i-lit case-literals :test #'equal)
	  (setf is-memberset nil))))))


(defun stepcase-memberset (store-properties prob-properties)
  (let ((is-memberset t))
    (dolist (i-prop prob-properties is-memberset)
      (when is-memberset
	(unless (find i-prop store-properties)
	  (setf is-memberset nil))))))


(defun realinst-predicate (pred-pos)
  (let* ((predicates (dom-predicates *pspace*))
	 (objects (problem-inheritobjects *current-problem*))
	 (poslist (objpos-from-mappos (cdr pred-pos) 
				      (gethash (car pred-pos) predicates) objects)))
	(cons (car pred-pos) (instances-from-poslist poslist objects))))



(defun object-typebitmap (object type objects)
  (let* ((objects-of-type (gethash type objects))
	 (bitmap (make-array (length objects-of-type) :element-type 'bit :initial-element 0)))
    (setf (sbit bitmap (position object objects-of-type)) 1)
    bitmap))


(defun pred-relevant-bitmaps (object args pos-relevant objects)
  (mapcar #'(lambda (i-arg)
	      (if (= pos-relevant (position i-arg args))
		  (object-typebitmap object (cdr i-arg) objects)
		(make-array (length (gethash (cdr i-arg) objects)) 
			    :element-type 'bit :initial-element 1)))
	  args))



(defun get-goals-objects ()
  (let ((inst-goals (get-instantiated-goals (problem-goals *current-problem*)))
	(goals-objects nil))
    (dolist (igoal inst-goals goals-objects)
      (let ((goal-obs (cdr (realinst-predicate igoal))))
	(dolist (i-ob goal-obs)
	  (unless (find i-ob goals-objects)
	    (push i-ob goals-objects)))))))


;;=============================================================
;; General Print Functions for CBR Module
;;=============================================================

(defun cbrpp-replay-table ()
  (pp-hash-table *replay-table*))
(defun cbrpp-replay-seq (object)
  (replay-sequence (gethash object *replay-table*)))
 
(defun cbrpp-typebase-field (cb-field)
  (maphash (lambda (type base)
	     (format t "~%~% Case Base Field ~a for ~a" cb-field type)
	     (maphash (lambda (numcase v_case)
			(format t "~% ~a ~a" numcase (funcall (symbol-function cb-field) v_case)))
		      base))
   *type-case-base*))
		      