;; SAYPHI Planner 
;; Planning-Domain Functions for Search Algorithms
;; Td. 16.12.2006
;; _________________________________________________________________________________________

;; Stores node pointer to acces them by number
(defvar *hash-nodes* (make-hash-table))
(defvar *hash-duplicates* nil)
(defvar *hash-inconsistent-h* nil)

(defparameter *applicable-byte* nil)


;;Duplicates the hash table containing the maps of literals
(defun copy-state (hash-state)
  (let ((new-state (make-hash-table :test #'eq)))
    (maphash #'(lambda (key value)
		 (setf (gethash key new-state) (copy-seq value))) 
	     hash-state)
    new-state))

;;Copy only the sets of bitmaps or the set of arrays-fluents
(defun copy-state-set (hash-state &optional (set 'facts))
  (let ((new-state (make-hash-table :test #'eq)))
    (maphash #'(lambda (key value)
		 (cond ((eq set 'facts)
			(when (bit-vector-p value)
			  (setf (gethash key new-state) (copy-seq value))))
		       ((eq set 'fluents)
			(unless (bit-vector-p value)
			  (setf (gethash key new-state) (copy-seq value))))
		       ((eq set 'real-fluents)
			(when (or (and (not (bit-vector-p value))
				       (find key (dom-real-functors *pspace*)))
				  (eq key 'total-time))
			  (setf (gethash key new-state) (copy-seq value))))
		       ))
			
	     hash-state)
    new-state))

(defun is-metric-domain ()
  (when (and (not (null *current-problem*))
	     (not (null (problem-metric *current-problem*)))
	     (find :fluents (dom-requirements *pspace*)))
	     t))





(defmacro is-relaxed-plan-member (relaxed-plan action)
  `(not (null (find ,action ,relaxed-plan :test #'equal :key #'car))))

;; It substitutes the snode slot after representation change
(defmacro snode-plan-action (node)
  `(when (gaction-p (snode-applied-action ,node))
    (gaction-planaction (snode-applied-action ,node))))


;; It computes the left goals of the current node, when we want to deal
;; only with not achieved goals 
(defun target-goals (node &optional (goals (problem-goals *current-problem*)))
  (let ((state (if (snode-p node) (snode-state node) node))
	(left-goals (make-hash-table)))
    (maphash #'(lambda (pred goal-bitmap)
		 (when (bit-vector-p goal-bitmap)
		   (setf (gethash pred left-goals)
			 (bit-andc2 goal-bitmap (gethash pred state)))))
	     goals)
    left-goals))



;;Determines if the goals are reached in the given state
;;I check the type because i do not create nodes at relax-plan
(defun goals-reached (node &optional (goals (problem-goals *current-problem*)))
  (let ((state (if (snode-p node) (snode-state node) node))
	(are-reached t))
    (maphash #'(lambda (pred goal-bitmap)
		 (when are-reached
		   (let ((state-bitmap (gethash pred state)))
		     (unless (equal goal-bitmap (bit-and state-bitmap goal-bitmap))
		       (setf are-reached nil)))))
	     goals)
    are-reached))


;;Creates a bitmap of a given literal goal.  
;;Use this only if you need to compute in bitmaps
(defun onegoal-bitmap (literal-goal)
  (let ((goal-map (make-hash-table :test #'eq))
	(pred-pos (fact-predpos (gethash literal-goal *facts*))))
    (setf (gethash (car pred-pos) goal-map)
	  (copy-seq (gethash (car pred-pos) (problem-patterns *current-problem*))))
    (change-state goal-map (car pred-pos) (cdr pred-pos) 1)
  goal-map
  ))


;; If state1 and state2 are the same
(defun equal-state (node1 node2)
  (let ((are-equal t)
	(state2 (snode-state node2)))
    (maphash #'(lambda (pred1 pred-bitmap1)
		 (when are-equal
		   (when (or (bit-vector-p pred-bitmap1)
			     (find pred1 (dom-positive-vars *pspace*))
			     (find pred1 (dom-negative-vars *pspace*)))
		     (unless (equalp pred-bitmap1 (gethash pred1 state2))
			 (setf are-equal nil)))))
	     (snode-state node1))
    are-equal
    ))



(defmacro apply-inst-effects (state effs bitval)
  `(dolist (i-eff ,effs)
     (change-state ,state (car i-eff) (cdr i-eff) ,bitval)))



(defun apply-action-instance (state action)
  (let ((new-state (copy-state state)))
    (apply-inst-effects new-state (gaction-dels action) 0)
    (apply-inst-effects new-state (gaction-adds action) 1)
     (when (is-metric-domain)
       (apply-gaction-costs action new-state))
    new-state))


(defun compute-duphash-code (state)
  (let* ((duphash-pred (get (problem-plist *current-problem*) :duphash-pred))
	 (random-matcher (get (problem-plist *current-problem*) :duphash-matcher))
	 (predicate-seq (subseq (gethash duphash-pred state) 0 (length random-matcher))))
    (bitmap-to-integer (bit-xor predicate-seq random-matcher))))



;;=============================================================================
;;Deciding if an action if applicable. 
;;If the relaxed-plangraph is called first in node, this is avoided using
;; the applicable of helpful byte of the node
(defun connect-fact (pred-pos)
  (let ((fg (fact-of-predpos pred-pos)))
    (dolist (iact (fg-precond-of fg))
      (unless (gaction-applicable iact)
	(incf (gaction-prec-gcount iact))
	(when (and (= (gaction-num-precs iact) (gaction-prec-gcount iact))
		   (= (gaction-num-precfuns iact) (gaction-precfun-gcount iact)))
	  (set-instance-map *applicable-byte* (gaction-int iact) 1)
	  (setf (gaction-applicable iact) t)
	  )))))      


(defun connect-fluent (pred-pos state-value)
  (let ((flg (fluent-of-predpos pred-pos)))
    (dolist (iop-precfun-n (flg-precfun-of flg))
      (let* ((iop (car iop-precfun-n))
	     (precfun (nth (cdr iop-precfun-n) (gaction-precfuns iop))))
	(unless (gaction-applicable iop)
	  (when (precfun-value-eval precfun state-value)
	    (incf (gaction-precfun-gcount iop))
	      (when (and (= (gaction-num-precs iop) (gaction-prec-gcount iop))
			 (= (gaction-num-precfuns iop) (gaction-precfun-gcount iop)))
		(setf (gaction-applicable-h iop) t)
		(set-instance-map *applicable-byte* (gaction-int iop) 1)
		)))))))




(defun reset-expand-state ()
  (setf *applicable-byte* 0)
  (dolist (iact *actions*)
    (setf (gaction-prec-gcount iact) 0)
    (setf (gaction-precfun-gcount iact) 0)
    (setf (gaction-applicable iact) nil)))



(defun find-applicable-byte (node helpful)
  (cond 
;;     ((and (eq helpful 'mix-parent) (integerp (snode-helpful-byte node)))
;; 	 (setf (snode-applicable-byte node)
;; 	       (logand (if (integerp (snode-applicable-byte node))
;; 			   (snode-applicable-byte node)
;; 			   (set-applicable-byte node))
;; 		       (logior (snode-helpful-byte node)
;; 			       (if (snode-p (snode-parent node))
;; 				   (snode-helpful-byte (snode-parent node)) 0)))))
	((and helpful (integerp (snode-helpful-byte node)))
	 (snode-helpful-byte node))
	((integerp (snode-applicable-byte node)) 
	 (snode-applicable-byte node))
	(t 
	  (setf (snode-applicable-byte node) (set-applicable-byte node)))))

(defun set-applicable-byte (node)
  (reset-expand-state)
  (maphash #'(lambda (key statemap)
	       (cond ((simple-bit-vector-p statemap)
		      (dotimes (ibit (length statemap))
			(when (= 1 (bit statemap ibit))
			  (connect-fact (cons key ibit)))))
		     (t
		      (dotimes (ibit (length statemap))
			(when (numberp (aref statemap ibit))
			  (connect-fluent (cons key ibit) (numvar-state-value key (snode-state node) ibit))
			  )))))
		    (snode-state node))
  *applicable-byte*)

;;=====================================================================================

(defun expand-state (node &key (helpful t))
  (let* ((state (snode-state node)) (children nil)
	 (applicable-byte (find-applicable-byte node helpful))
	 (new-child nil) (child-node nil)) 
    (dotimes (ibit (integer-length applicable-byte))
      (when (logbitp ibit applicable-byte)
	(let ((app-action (aref *actions-table* ibit)))
	  (setf new-child (apply-action-instance state app-action))
	  (setf (snode-expanded node) t)
	  (setf child-node (make-snode 
			  :depth (+ 1 (snode-depth node))
			  :parent node
			  :state new-child
			  :applied-action app-action
			  :helpful-p helpful
			  ))
 	  (setf (snode-cost child-node) 
		(action-cost-metric state new-child))
	  (setf (snode-number child-node) 
		(incf (get (problem-plist *current-problem*) :node-counter)))
	  (setf (gethash (snode-number child-node) *hash-nodes*) child-node)
	  (setf (snode-hash-code child-node) (compute-duphash-code new-child))
	  (push child-node children))
	      ))
    (setf (snode-children node) (reverse children))
))
  


(defun get-instantiated-goals (goals)
  (let ((inst-goals nil))
    (maphash #'(lambda (pred goals-bitmap)
		 (dotimes (ibit (length goals-bitmap))
		   (when (eq 1 (sbit goals-bitmap ibit))
		     (push (cons pred ibit) inst-goals))))
	     goals)
    inst-goals))


;; Check this out first. It may not work after the new representation
;; Searchs for all applicable actions and stores candidates nodes for later creation
;; (defun expand-candidate-state (node)
;;   (let ((actions (dom-actions *pspace*))
;; 	(state (snode-state node)))
;;     (setf (snode-candidates node) nil)
;;     (dolist (iaction actions (snode-children node))
;;       (let ((applicable-map (action-applicable-byte iaction state)))
;; 	(when (> applicable-map 0)
;; 	  (dotimes (ibit (integer-length applicable-map))
;; 	    (when (logbitp ibit applicable-map)
;; 	      (setf (snode-expanded node) t)
;;  	      (setf (snode-candidates node)
;;  		    (append (snode-candidates node) (list (cons (action-name iaction) ibit))))
;; 	      )))))))
;; 
;; 
;; ;; Creates a node's child from one of its candidates	      
;; (defun create-candidate-node (node &optional (problem *current-problem*))
;;   (let ((candidate (pop (snode-candidates node)))
;; 	(new-child nil) (child-node nil))
;;     (when candidate
;;       (let ((iaction (gethash (car candidate) (get (dom-plist *pspace*) :direct-actions)))
;; 	    (state (snode-state node)))
;; 	(setf new-child (apply-action-instance state iaction (cdr candidate)))
;; 	(setf child-node (make-snode 
;; 			  :number (incf (get (problem-plist problem) :node-counter))
;; 			  :depth (+ 1 (snode-depth node))
;; 			  :parent node
;; 			  :state new-child
;; 			  :applied-action candidate))
;; 	(push child-node (snode-children node))))))
  




   






