;; SAYPHI Planner 
;; Forward Guided Search Algorithms
;; Td. 16.12.2006
;; _________________________________________________________________________________________


(defvar *trace-search* nil) 
(defvar *trace-search-depth* 3)
(defvar *trace-node* nil) 
(defvar *trace-listnodes* (make-hash-table))
(defvar *trace-ehc* nil)


;; This are global variables for defaults planning's parameters
(defvar *say-algorithm* 'enforced-hill-climbing)
(defvar *say-heuristic* 'h-metric-rxplan)
(defvar *say-costfn* 'node-path-cost)
(defvar *say-helpful* t)
(defvar *say-depthbound* 1000)
(defvar *say-timeout* 30)
(defvar *say-h-rounded* t)

;; 0:silent 1:plan 2:search & plan
(defvar *say-solution* nil)
(defvar *search-failed* nil)
(defvar *say-2try* t) ;; If first algorithm fails, try a complete best-first

;;tracing CBPlanning
(defvar *hit-node* nil)
(defvar *hit-average* 0)

(defun reset-ehc-trace()
  (setf *hit-node* nil
	*hit-average* 0))


(defun build-path (node)
  (let ((plan-path nil))
    (do ((i-node node (snode-parent i-node)))
	((null (snode-parent i-node)) plan-path)
	(setf (snode-selected (snode-parent i-node)) i-node)
	(push i-node plan-path))))


(defun reach-time-bound (problem)
  (let* ((initial-search-time (get (problem-plist problem) :initial-search-time))
	 (current-time (get-internal-run-time))
	 (searching-seconds (float (/ (- current-time initial-search-time)
				      internal-time-units-per-second))))
    (when (> searching-seconds (get-sayp :say-timeout)) t)))



(defun build-solution (node problem found reason)
  (let* ((stop-time (get-internal-run-time))
	 (initial-pre-time (get (problem-plist problem) :initial-pre-time))
	 (initial-search-time (get (problem-plist problem) :initial-search-time))
	 (this-solution (make-solution 
			 :found found
			 :total-time (float (/ (- stop-time initial-pre-time)
					       internal-time-units-per-second))
			 :pre-time (float (/ (- initial-search-time initial-pre-time)
					     internal-time-units-per-second))
			 :search-time (float (/ (- stop-time initial-search-time)
						internal-time-units-per-second))
			 :depth (when (snode-p node) (snode-depth node))
			 :path (when (snode-p node) (build-path node))
			 :num-nodes (1+ (get (problem-plist problem) :node-counter)) ;;mas el nodo inicial
			 :evaluated-nodes (get (problem-plist problem) :node-evaluated)
			 :last-node node
			 :stop-reason reason
			 :total-cost (compute-total-cost node problem))))
    this-solution))



;;  I could put this in a way of signals handlers to add more functions to check each cycle
(defun stop-search (node problem)
  (cond ((null node)
	 (cons t :search-failed))
	((= (snode-h-value node) most-positive-fixnum)
 	 (cons t :recognized-dead-end))
	((goals-reached node)
	 (cons t :goals-reached))
	((>= (snode-depth node) (get-sayp :say-depthbound))
	 (cons t :depth-bound))
	((reach-time-bound problem)
	 (cons t :time-bound))
	((and *trace-search* (>= (snode-depth node) *trace-search-depth*))
	 (cons t :trace-search))
	(t nil)))


(defun when-stop (reason node problem)
  (case reason
	(:goals-reached
	 (setf *say-solution* (build-solution node problem t reason))
	 (say-pp-solution *say-solution*))
	(:recognized-dead-end
	 (format t "~% DEAD END - UNREACHABLE GOALS")
	 (setf *say-solution* (build-solution node problem nil reason)))
	(:depth-bound 
	 (format t "~% HIT DEPTH BOUND")
	 (build-solution node problem nil reason))
	(:time-bound
	 (format t "~% HIT TIME BOUND")
	 (build-solution node problem nil reason))
	(:search-failed 
	 (format t "~% SEARCH FAILED!!")
	 (setf *search-failed* t)
	 (build-solution node problem nil reason))
	(:trace-search
	 (build-solution node problem nil reason)
	 (format t "~% Trace search at depth ~d !!" (snode-depth node))
	 (setf *trace-node* node))))





(defun hc-next-node (node visited sort-function)
  (stable-sort (snode-children node) sort-function :key #'snode-h-value)
  (let ((selected-node nil) (found-next nil))
    (dolist (i-child (snode-children node) selected-node)
      (unless found-next
	(unless (find i-child visited :test #'equal-state)
	  (setf selected-node i-child)
	  (setf found-next t))))))


;; Select the first open from chlis or backtrack to the next open node
(defun next-node (node sort-function)
  (let ((node-selected nil))
    (cond ((snode-p (setf node-selected 
			  (find-if #'(lambda (xnode)
				       (and (null (snode-closed xnode))
					    (funcall sort-function (snode-h-value xnode) 
						     (snode-h-value node))))
				   (snode-children node))))
	   node-selected)
	  (t
	   (setf (snode-closed node) t)
	   (cond ((not (null (snode-parent node)))
		  (next-node (snode-parent node) sort-function))
		 (t nil))))))


(defun next-open (node)
  (let ((node-selected nil))
    (cond ((snode-p (setf node-selected 
			  (find-if #'(lambda (xnode)
				       (null (snode-closed xnode)))
				   (snode-children node))))
	   node-selected)
	  (t
	   (setf (snode-closed node) t)
	   (cond ((not (null (snode-parent node)))
		  (next-open (snode-parent node)))
		 (t nil))))))



(defun map-heuristic (node h-function)
  (mapc #'(lambda (v-node) 
	    (multiple-value-bind 
	     (h-value relaxed-plan focus-goals) (funcall h-function v-node)
	     (setf (snode-h-value v-node) h-value)
	     (setf (snode-relaxed-plan v-node) relaxed-plan)
	     (setf (snode-focus-goals v-node) focus-goals)
	     (incf (get (problem-plist *current-problem*) :node-evaluated))
	     ))
	(snode-children node)))


;; It's like map-heuristic, but for an especific node
(defun store-h-extras (node h-function)
  (multiple-value-bind 
   (h-value relaxed-plan focus-goals) (funcall h-function node)
     (setf (snode-h-value node) h-value)
     (setf (snode-h-plus node) (length relaxed-plan))
     (setf (snode-relaxed-plan node) relaxed-plan)
     (setf (snode-focus-goals node) focus-goals)
     (incf (get (problem-plist *current-problem*) :node-evaluated))
     ))
  
(defun store-a-star-extras (inode cost-fn)
  (setf (snode-g-value inode) (funcall cost-fn inode))
  (setf (snode-f-value inode) (+ (snode-h-value inode) (snode-g-value inode))))


(defun remove-node-by-number (parent node-number)
  (remove-if #'(lambda (ichild)
		 (when (= (snode-number ichild) node-number) t))
	     (snode-children parent)))


(defun map-astar-f (nodes cost-fn h-fn)
  (mapc #'(lambda (inode) 
	    (store-h-extras inode h-fn)
	    (setf (snode-g-value inode) (funcall cost-fn inode))
	    (setf (snode-f-value inode) (+ (snode-h-value inode) (snode-g-value inode))))
	nodes))


(defun node-path-cost (node)
  (cond ((snode-p (snode-parent node))
	 (+ (snode-g-value (snode-parent node)) (snode-cost node)))
	(t 0)))



(defun restore-nonhelpful (node h-fn)
  (setf (snode-children node) nil)
  (expand-state node :helpful nil)
  
  (map-heuristic node h-fn)
  (snode-children node)
)


(defun lessthan-f-metricplus (node1 node2)
  (cond ((< (snode-f-value node1) (snode-f-value node2)) t)
	((and (= (snode-f-value node1) (snode-f-value node2))
	      (< (snode-h-plus node1) (snode-h-plus node2))) t)
	(t nil)))


(defun recompute-relinked (node cost-fn)
  (dolist (jchild (snode-children node))
    (unless (snode-closed jchild)
      (setf (snode-depth jchild) (1+ (snode-depth (snode-parent jchild))))
      (store-a-star-extras jchild cost-fn)
      (unless (null (snode-children jchild))
	(recompute-relinked jchild cost-fn)))))


(defun relink-g-parent (node old-repeated cost-fn)
  (let ((parent (snode-parent node))
	(old-parent (snode-parent old-repeated)))
    (when (= 2 *say-output*)
      (format t "~% ============RELINKING ============================== ")
      (format t "~% New Node: ~a          Parent    : ~a" (snode-number node) (snode-number parent))
      (format t "~% Repeated: ~a          Old Parent: ~a" (snode-number old-repeated) (snode-number old-parent))
      (format t "~% ============RELINKING ============================== "))
    
    (setf (snode-closed node) t)
    
    (setf (snode-parent old-repeated) parent)
    (setf (snode-depth old-repeated) (1+ (snode-depth parent)))
    
;;;;;;;;;;     (setf (snode-children parent) (remove-node-by-number parent (snode-number node)))
    (push old-repeated  (snode-children parent))
    (setf (snode-children old-parent) (remove-node-by-number old-parent (snode-number old-repeated)))
    (recompute-relinked old-repeated cost-fn)
    ))




(defun better-g-path (repeated-node node cost-fn)
  (let ((new-g (funcall cost-fn node)))
    (cond ((< new-g (snode-g-value repeated-node)) t))))



(defun get-enforced-heuristic (node h-function current-h visited-nodes &key (discrete nil))
  (let ((better-found nil) (node-counter 0))
    (dolist (i-child (snode-children node) better-found)
;;       (format t "~% >> Evaluating NODE ~a ~%" i-child)
      (when (null better-found)
	(cond ((not (find i-child visited-nodes :test #'equal-state))
;; 	       (format t "~% >>>> Evaluating...")
	       (incf node-counter)
	       (store-h-extras i-child h-function)
	       (cond (discrete
		      (when (< (snode-h-value i-child) current-h)
			(setf better-found i-child)
			(when *trace-ehc*
			  (push node-counter *hit-node*))))
		     (t
		       (when (or (< (snode-h-value i-child) current-h)
				 (and (= (snode-h-value i-child) current-h)
				      (< (snode-h-plus i-child) (snode-h-plus (snode-parent i-child)))))
			 (setf better-found i-child))))) 
	      (t
;; 	       (format t "~% >>>> Closing...")
	       (setf (snode-closed i-child) t)))
	       ))))


(defun lessthan-h-metricplus (node1 node2)
  (cond ((< (snode-h-value node1) (snode-h-value node2)) t)
	((and (= (snode-h-value node1) (snode-h-value node2))
	      (< (snode-h-plus node1) (snode-h-plus node2))) t)
	(t nil)))


(defun ehc-open-nodes (open-children sort-option)
  (cond ((equal sort-option 'h-value)
	 (stable-sort open-children #'lessthan-h-metricplus))
	(t open-children)))






(defun enforced-hill-climbing (init-node h-fn cost-fn search-options &optional (problem *current-problem*))
  (declare (ignore cost-fn))
  (let ((open-nodes nil) (open-children nil)
	(node init-node) (next-node nil) (visited nil)
	(current-h 0) (better-h-node nil))

;;     (if *generate-learning-examples-p*
;; 	(setf *examples* nil))
    (store-h-extras node h-fn)
    (setf current-h (snode-h-value node))
;;     (setf open-nodes (list node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))
      
      (expand-state node :helpful *say-helpful*)
      (print-search-node node nil)
	     
      (push node visited)
      
      (setf better-h-node (get-enforced-heuristic node h-fn current-h visited))
      (if *generate-learning-examples-p*
	  (save-examples node better-h-node nil))
      (cond ((snode-p better-h-node)
	     (setf current-h (snode-h-value better-h-node))
	     (setf node better-h-node)
	     (setf open-nodes nil)
	     (setf open-children nil))
	    (t 
	     (setf open-children (append (remove-if #'snode-closed (snode-children node)) 
				       open-children))
	     
	     (cond ((not (snode-p (setf next-node (pop open-nodes))))
		    (when (> *say-output* 1) (format t "  ~% [Expanding Breadth Level]"))
		    (setf open-nodes (ehc-open-nodes open-children (car (find-argument search-options :children-sort))))
		    
		    (when (null open-nodes) ;;Recuperando la poda por las helpful actions
		      (setf open-nodes (restore-nonhelpful node h-fn))
		      (when (> *say-output* 1) 
			(format t "  ~%!!Restoring NON Helpful Nodes.")))
		    (setf node (pop open-nodes))
		    (setf open-children nil))
		   (t 
		    (setf node next-node)))))       
      ;; 	 (when (snode-p node)
      ;; 	   (print-search-node node nil)
      )))

 
		
(defun a-star (init-node h-fn cost-fn search-options)
  (declare (ignore search-options))
  (let ((open-nodes nil) (node init-node) (repeated nil) (visited nil) (current-h 0)
	(problem *current-problem*))
    (store-h-extras node h-fn)
    (store-a-star-extras node cost-fn)
    (setf current-h (snode-h-value node))
    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))

      (setf repeated nil)
      (push node visited)
      (expand-state node :helpful nil)
      (dolist (i-child (snode-children node))
	(setf repeated (find i-child visited :test #'equal-state))
       	(cond ((snode-p repeated)
;; 	       (format t "~% ========?? Repeated visited ~a" i-child)
	       (cond ((better-g-path repeated i-child cost-fn)
		      (relink-g-parent i-child repeated cost-fn))
		     (t 
		      (setf (snode-closed i-child) t)))))
	
	(setf repeated (find i-child open-nodes :test #'equal-state))
	(cond ((snode-p repeated)
;; 	       (format t "~% ========?? Repeated open-list ~a" i-child)
	       (cond ((better-g-path repeated i-child cost-fn)
		      (setf (snode-closed repeated) t))
		     (t 
		      (setf (snode-closed i-child) t)))))
       
	(unless (snode-closed i-child)
	  (store-h-extras i-child h-fn)
	  (store-a-star-extras i-child cost-fn)
	  (push i-child open-nodes)))
	
      (setf open-nodes (stable-sort (remove-if #'snode-closed open-nodes) 
				    #'< :key #'snode-f-value))
      
      (setf node (pop open-nodes))
      
      (when (snode-p node)
	(print-search-node node open-nodes))
      )))




(defun hill-climbing (init-node h-function cost-fn &key (minimize-h t) 
				(problem *current-problem*)
				(helpful *say-helpful*))
  (declare (ignore cost-fn))
  (let ((sort-function (if minimize-h #'< #'>))
	(visited nil)
	(node init-node)
	(next-node nil))
    (store-h-extras node h-function)

    (do* ((stop-this (stop-search node problem) (stop-search node problem)))
	 (stop-this (when-stop (cdr stop-this) node problem))
	 (expand-state node :helpful helpful)
	 (map-heuristic node h-function)

	 (setf next-node (hc-next-node node visited sort-function))
	 (cond ((snode-p next-node)
		(push node visited)
		(setf node next-node))
	       (t
		(restore-nonhelpful node h-function)
		(setf next-node (hc-next-node node visited sort-function))
		(when (snode-p next-node)
		  (push node visited))
		(setf node next-node)))
	 (when (snode-p node)
	   (print-search-node node nil))
;; 	 (break)
	 )))

  

      


(defun initialize-current-problem ()
  (let ((problem *current-problem*))
    (update-artificial-initstate)
    (build-actions-instpattern)

    (build-actions-bytes)
    (build-direct-predinstance)
    
    (setf (dom-pred-achievers *pspace*) (build-pred-achievers))
    (setf (dom-numeric-achievers *pspace*) (build-metric-achievers))
    (setf (get (dom-plist *pspace*) :direct-actions) (create-actions-direct-access))
        
    (create-inst-precfuns)
    (create-inst-costs)
    (split-numvars-range)
    (remove-redundant-inst-action)
    
    (setf (get (problem-plist problem) :node-counter) 0)
    (setf (get (problem-plist problem) :node-evaluated) 0)
    (setf (get (problem-plist problem) :inst-goals) 
	  (get-instantiated-goals (problem-goals problem)))
    (setf (problem-search-tree problem)
	  (make-snode :number 0
		      :depth 0
		      :state (copy-state (problem-init-state problem))))))



(defun say-plan-defaults (plan-option)
  (case plan-option
    (heuristic (cond ((is-metric-domain) #'h-metric-rxplan)
		     (t #'h-relaxedplan)))
    (algorithm #'enforced-hill-climbing)))


(defun plan (&key (algorithm nil)
		  (heuristic nil)
		  (cost *say-costfn*)
	          (runtype nil)
		  (timeout *say-timeout*)
		  (depthbound *say-depthbound*)
		  (try2-search *say-2try*))
  (declare (special *current-problem*)
	   (ignore try2-search))
  (let ((problem *current-problem*)
	(i-node nil)
	(algorithm-fn (if (not (null algorithm)) (symbol-function algorithm) (say-plan-defaults 'algorithm)))
	(heuristic-fn (if (not (null heuristic)) (symbol-function heuristic) (say-plan-defaults 'heuristic)))
	(cost-fn (symbol-function cost))
	(start-time (get-internal-run-time))
	(search-options nil)
	(sol nil))


    (set-sayp :say-timeout timeout)
    (set-sayp :say-depthbound depthbound)

    (reset-ehc-trace)
    (sayout-initialize)
    (setf i-node (initialize-current-problem))
    
    (setf (get (problem-plist problem) :initial-pre-time) start-time)
    (setf (get (problem-plist problem) :initial-search-time) (get-internal-run-time))
    
    (setf  *search-failed* nil)
    (setf *say-solution* nil)

    (case runtype
      (ehc-sort-h
       (setf search-options '((:children-sort h-value)))))
    
    (sayout-search t algorithm heuristic cost)
    (setf sol (funcall algorithm-fn i-node heuristic-fn cost-fn search-options))

;;     Turning to best-first algorithm without helpful actions
;;     (cond ((and *search-failed* try2-search)
;; 	   (setf i-node (initialize-current-problem))
;; 	   (setf sol (best-first i-node heuristic-fn cost-fn)))
;; 	  (t sol))

))
